

class Axis:

    def __init__(self, data, label: str, color: str = '#0074D9', linestyle: str = 'solid'):
        self.__data = data
        self.__label = label
        self.__color = color
        self.__linestyle = linestyle

    @property
    def data(self):
        return self.__data

    @property
    def label(self):
        return self.__label

    @property
    def color(self):
        return self.__color

    @property
    def linestyle(self):
        return self.__linestyle


from simses.commons.profile.file import FileProfile
from simses.commons.config.simulation.general import GeneralSimulationConfig
from simses.commons.config.simulation.profile import ProfileConfig
from simses.commons.profile.technical.technical import TechnicalProfile


class FrequencyProfile(TechnicalProfile):

    def __init__(self, config: GeneralSimulationConfig, profile_config: ProfileConfig):
        super().__init__()
        self.__file: FileProfile = FileProfile(config, profile_config.frequency_file)
        #self.__file = open(profile_config.frequency_file, 'r', newline='')
        #self.__header_missing_exception = False
        #self.__first_line = None
        #self.__header_length = None
        #self.__header = self.get_header()
        # TODO Fix imported get header function:
        # self.__header = get_header_from(profile_config.frequency_file)

    def next(self, time) -> float:
        # Handling of profile files with and without header
        #if self.__header_missing_exception is True:
        #    __frequency = self.__first_line
        #    self.__header_missing_exception = False
        #else:
        #    __frequency = self.__file.readline()
        #return float(__frequency)
        return self.__file.next(time)

    #def get_header(self) -> dict:
    #   """
    #   Analyzes the header structure and saves it into a dict.
    #   Furthermore sets the __header_length parameter.
    #
    #   Returns
    #   -------
    #   dict
    #       Containing the header parameters and their values.
    #   """
    #
    #   __header = {}
    #   __line = self.__file.readline()
    #   __line_count = 1
    #   if __line not in ['"""\r\n']:
    #       self.__header_missing_exception = True
    #       self.__first_line = __line
    #       __header = None
    #   else:
    #       while __line not in ['\r\n']:
    #           if '#' in __line:
    #               key_raw, entry_raw = __line.split(sep=':', maxsplit=1)
    #               key = key_raw.strip('# ')
    #               entry = entry_raw.strip()
    #               __header[key] = entry
    #           __line = self.__file.readline()
    #           __line_count += 1
    #       self.__header_length = __line_count
    #   return __header

    def close(self):
        self.__file.close()



from abc import ABC, abstractmethod
from simses.system.auxiliary.auxiliary import Auxiliary


class HeatingVentilationAirConditioning(Auxiliary, ABC):

    def __init__(self):
        super().__init__()

    @abstractmethod
    # temperature_timestep: time
    def run_air_conditioning(self, temperature_time_series: [float], temperature_timestep: float, ambient_air_temperature: float) -> None:
        pass

    @abstractmethod
    def get_electric_power(self) -> float:
        pass

    @abstractmethod
    def get_max_thermal_power(self) -> float:
        pass

    @abstractmethod
    def get_thermal_power(self) -> float:
        pass

    @abstractmethod
    def get_set_point_temperature(self) -> float:
        pass

    @abstractmethod
    def get_scop(self) -> float:
        pass

    @abstractmethod
    def get_seer(self) -> float:
        pass

    @abstractmethod
    def get_temperature_dead_band(self) -> float:
        pass

    @abstractmethod
    def update_air_parameters(self, air_mass: float = None, air_specific_heat: float = None, air_density: float = None) -> None:
        pass

    def get_coefficients(self) -> [float, float, float]:
        """
        Needed for plotting BA Hörmann
        """
        pass

    def get_plotting_arrays(self) -> [[float], [float], [float], [float], [float], [float]]:
        """
        Needed for plotting BA Hörmann
        """
        pass

    def ac_operation_losses(self) -> float:
        return self.get_electric_power()

    @abstractmethod
    def set_electric_power(self, electric_power: float) -> None:
        pass



from simses.commons.state.technology.lithium_ion import LithiumIonState
from simses.technology.lithium_ion.cell.type import CellType
from simses.technology.lithium_ion.degradation.calendar.calendar_degradation import CalendarDegradationModel


class NCAGenericCalendarDegradationModel(CalendarDegradationModel):

    def __init__(self, cell_type: CellType):
        super().__init__(cell_type)
        self.__capacity_loss = 0
        self.__total_time_passed = 0
        self.__initial_capacity = self._cell.get_nominal_capacity()

        # parameters
        self.__z = 0.409609637073964
        self.__a = 0.00461276336758222
        self.__b = 15.2653523010812
        self.__c = 0.119518828585308
        self.__d = 0.0639038983833157
        self.__e = 1
        self.__f = 1

    def time_dependency(self, time_passed):
        # degradation depends on time as a power law function
        return time_passed ** self.__z

    def temperature_dependency(self, temp):
        # degradation depends on temperature as a power law function
        return self.__a * (temp / 273.15) ** self.__b

    def soc_dependency(self, soc):
        # degradation depends on soc as a linear function
        return self.__c * soc + self.__d

    def calculate_degradation(self, time: float, battery_state: LithiumIonState) -> None:
        self.__total_time_passed += time - battery_state.time
        timestep = time - battery_state.time
        state_of_charge: float = battery_state.soc
        temperature: float = battery_state.temperature

        # calculate degradation in percent (divide by 100 to get p.u.) with function 5 from MA Sandner
        capacity_loss_after = self.time_dependency(self.__total_time_passed) * \
                              self.temperature_dependency(temperature) * \
                              self.soc_dependency(state_of_charge)
        capacity_loss_before = self.time_dependency(self.__total_time_passed - timestep) * \
                              self.temperature_dependency(temperature) * \
                              self.soc_dependency(state_of_charge)

        self.__capacity_loss = (capacity_loss_after - capacity_loss_before) / 100.0 * self.__initial_capacity

    def calculate_resistance_increase(self, time: float, battery_state: LithiumIonState) -> None:
        # no resistance increase calculated for generic degradation model
        pass

    def get_degradation(self) -> float:
        degradation = self.__capacity_loss
        self.__capacity_loss = 0
        return degradation

    def get_resistance_increase(self) -> float:
        # no resistance increase calculated for generic degradation model
        return 0

    def reset(self, battery_state: LithiumIonState) -> None:
        self.__capacity_loss = 0
        self.__total_time_passed = 0

    def close(self) -> None:
        pass



class ElectricalCellProperties:

    """
    ElectricalCellProperties inherits all necessary data for describing the electrical behaviour of a cell.
    """

    def __init__(self, cell_voltage: float, cell_capacity: float, min_voltage: float, max_voltage: float,
                 max_charge_rate: float, max_discharge_rate: float, self_discharge_rate: float,
                 coulomb_efficiency: float):
        """
        Constructor of ElectricalCellProperties

        Parameters
        ----------
        cell_voltage :
            nominal voltage of one cell in V
        cell_capacity :
            nominal capacity of one cell in V
        min_voltage :
            minimum allowed voltage of one cell in V
        max_voltage :
            maximum allowed voltage of one cell in V
        max_charge_rate :
            maximum allowed charge rate in 1/h (C-rate)
        max_discharge_rate :
            maximum allowed discharge rate in 1/h (C-rate)
        self_discharge_rate :
            self discharge rate in p.u. as X.X%-soc per day, e.g., 0.015 for 1.5% SOC loss per day
        coulomb_efficiency :
            coulomb efficiency of the cell in p.u.
        """
        self.__NOMINAL_VOLTAGE: float = cell_voltage  # V
        self.__NOMINAL_CAPACITY: float = cell_capacity  # Ah
        self.__MAX_VOLTAGE: float = max_voltage  # V
        self.__MIN_VOLTAGE: float = min_voltage  # V
        self.__MAX_CHARGE_RATE: float = max_charge_rate  # 1/h
        self.__MAX_DISCHARGE_RATE: float = max_discharge_rate  # 1/h
        self.__SELF_DISCHARGE_RATE: float = self_discharge_rate / 24.0 / 3600.0 / 100.0 # X.Xp.u-soc per second
        self.__COULOMB_EFFICIENCY: float = coulomb_efficiency  # p.u.

    def get_nominal_voltage(self) -> float:
        """

        Returns
        -------
        float:
            nominal voltage of one cell in V
        """
        return self.__NOMINAL_VOLTAGE

    def get_nominal_capacity(self) -> float:
        """

        Returns
        -------
        float:
            nominal capacity of one cell in Ah
        """
        return self.__NOMINAL_CAPACITY

    def get_max_voltage(self) -> float:
        """

        Returns
        -------
        float:
            maximum allowed voltage of one cell in V
        """
        return self.__MAX_VOLTAGE

    def get_min_voltage(self) -> float:
        """

        Returns
        -------
        float:
            minimum allowed voltage of one cell in V
        """
        return self.__MIN_VOLTAGE

    def get_max_charge_rate(self) -> float:
        """

        Returns
        -------
        float:
            maximum allowed charge rate in 1/h (C-rate)
        """
        return self.__MAX_CHARGE_RATE

    def get_max_discharge_rate(self) -> float:
        """

        Returns
        -------
        float:
            maximum allowed discharge rate in 1/h (C-rate)
        """
        return self.__MAX_DISCHARGE_RATE

    def get_self_discharge_rate(self) -> float:
        """

        Returns
        -------
        float:
            self discharge rate in Wh/s
        """
        return self.__SELF_DISCHARGE_RATE * self.get_nominal_capacity() * self.get_nominal_voltage()

    def get_coulomb_efficiency(self) -> float:
        """

        Returns
        -------
        float:
            coulomb efficiency of the cell in p.u.
        """
        return self.__COULOMB_EFFICIENCY



import pytest

from simses.system.auxiliary.pump.fixeta_centrifugal import FixEtaCentrifugalPump


@pytest.fixture(scope='function')
def uut(eta_pump):
    return FixEtaCentrifugalPump(eta_pump)


@pytest.mark.parametrize('eta_pump, pressure_loss, result', [(0.5, 100, 200)])
def test_get_pump(uut, pressure_loss, result):
    uut.calculate_pump_power(pressure_loss)
    assert uut.get_pump_power() == result



import pandas as pd
import scipy.interpolate
from math import log10

from simses.commons.config.data.redox_flow import RedoxFlowDataConfig
from simses.commons.config.simulation.redox_flow import RedoxFlowConfig
from simses.commons.log import Logger
from simses.commons.state.technology.redox_flow import RedoxFlowState
from simses.technology.redox_flow.stack.electrolyte.abstract_electrolyte import ElectrolyteSystem
from simses.technology.redox_flow.stack.abstract_stack import StackModule


class CellDataStack5500W(StackModule):
    """CellDataStack5500W is a stack based on experimental data of single cell measurements and scaled up to a 40-cell
    stack."""

    """The nominal power is calculated for an SOC range from 20 % to 80 % and the listed parameters. It corresponds to 
        the power that can be obtained over the entire SOC range.  """
    __STACK_POWER_NOM = 5500  # W
    __CELL_NUMBER = 40  # -
    __CELL_AREA = 2160  # cm^2
    __ELECTRODE_THICKNESS = 0.48  # cm
    __ELECTRODE_POROSITY = 0.9  # -
    __MIN_CELL_VOLTAGE = 1.2  # V
    __MAX_CELL_VOLTAGE = 1.6  # V
    __SELF_DISCHARGE_CURRENT_DENS_MEMBRANE = 0.000917  # A/cm^2
    __HYDRAULIC_RESISTANCE = 4.67e10  # 1/m^3
    __DEPENDENT_PARAMETERS = False

    __SOC_IDX = 0
    __TEMP_IDX = 1
    __2D_RINT = False  # True if temperature dependency should be considered in internal resistance.

    def __init__(self, electrolyte_system: ElectrolyteSystem, voltage: float, power: float,
                 redox_flow_data_config: RedoxFlowDataConfig, redox_flow_config: RedoxFlowConfig):
        super().__init__(electrolyte_system, voltage, power, self.__CELL_NUMBER, self.__STACK_POWER_NOM,
                         redox_flow_config)
        self.__log: Logger = Logger(__name__)
        self.__electrolyte_system: ElectrolyteSystem = electrolyte_system
        self.__RINT_FILE = redox_flow_data_config.rfb_rint_file_cell_stack
        self.__SHUNT_FILE = redox_flow_data_config.cell_stack_shunt_current
        self.__shunt_current = pd.read_csv(self.__SHUNT_FILE, header=None)  # A
        self.__shunt_current_mat = self.__shunt_current.iloc[1:, 1:]
        self.__soc_arr = self.__shunt_current.iloc[1:, 0]
        self.__current_arr = self.__shunt_current.iloc[0, 1:]
        self.__shunt_interp2d = scipy.interpolate.interp2d(self.__current_arr, self.__soc_arr, self.__shunt_current_mat,
                                                           kind='linear')
        self.__internal_resistance = pd.read_csv(self.__RINT_FILE)  # Ohm
        self.__soc_rint_arr = self.__internal_resistance.iloc[:, self.__SOC_IDX]
        self.__temp_rint_arr = self.__internal_resistance.iloc[:3, self.__TEMP_IDX]
        if self.__2D_RINT:
            self.__rint_mat_ch = self.__internal_resistance.iloc[:, 2:5]
            self.__rint_mat_dch = self.__internal_resistance.iloc[:, 5:]
            self.__rint_ch_inter2d = scipy.interpolate.interp2d(self.__temp_rint_arr, self.__soc_rint_arr,
                                                                self.__rint_mat_ch, kind='linear', fill_value=None)
            self.__rint_dis_inter2d = scipy.interpolate.interp2d(self.__temp_rint_arr, self.__soc_rint_arr,
                                                                 self.__rint_mat_dch, kind='linear', fill_value=None)
        else:
            self.__rint_mat_ch = self.__internal_resistance.iloc[:, 4]
            self.__rint_mat_dch = self.__internal_resistance.iloc[:, 7]
            self.__rint_ch_inter1d = scipy.interpolate.interp1d(self.__soc_rint_arr, self.__rint_mat_ch, kind='linear',
                                                                fill_value='extrapolate')
            self.__rint_dis_inter1d = scipy.interpolate.interp1d(self.__soc_rint_arr, self.__rint_mat_dch, kind='linear',
                                                                 fill_value='extrapolate')

    def get_open_circuit_voltage(self, redox_flow_state: RedoxFlowState) -> float:
        """
        Literature source: Fink, Holger. Untersuchung von Verlustmechanismen in Vanadium-Flussbatterien. Diss.
        Technische Universität München, 2019.
        equation 5.18, assumption: SOH = 100 %, therefore ver = 0.5
        """
        concentration_v = self.__electrolyte_system.get_vanadium_concentration()
        soc_stack = redox_flow_state.soc_stack
        # equation is defined for SOC = ]0-1[
        if soc_stack > 0.999:
            soc_stack = 0.999
        if soc_stack < 0.001:
            soc_stack = 0.001
        temperature = redox_flow_state.temperature
        concentration_h_start = 2.6  # mol/l
        ocv_cell = (1.255 + 0.07 + 0.059 * temperature / 298.15 * log10((soc_stack / (1 - soc_stack) *
                    (concentration_h_start + concentration_v / 1000 * (soc_stack + 0.5)))**2 * (concentration_h_start +
                    concentration_v / 1000 * (soc_stack - 0.5))))
        self.__log.debug('OCV_cell: ' + str(ocv_cell) + ' V')
        return ocv_cell * self.get_cell_per_stack() * self.get_serial_scale()

    def get_nominal_voltage_cell(self) -> float:
        """Calculated for a temperature of 25 °C and at SOC 50 %."""
        nominal_voltage_cell = 1.423
        return nominal_voltage_cell

    def get_internal_resistance(self, redox_flow_state: RedoxFlowState) -> float:
        """
        Experimental data based on single cell measurements at ZAE Bayern.
        Source: SimSES: Möller, Marc et al. A holistic simulation framework for modelingand analyzing stationary energy
        storage systems.
        """
        soc = redox_flow_state.soc
        temp = redox_flow_state.temperature
        if redox_flow_state.is_charge:
            # internal resistance for charging
            if self.__2D_RINT:
                resistance = self.__rint_ch_inter2d(temp, soc)
            else:
                resistance = self.__rint_ch_inter1d(soc)
            self.__log.debug('Resistance charging stack: ' + str(resistance) + ' Ohm, resistance charging module:'
                             + str(resistance / self.get_parallel_scale() * self.get_serial_scale()))
            return float(resistance / self.get_parallel_scale() * self.get_serial_scale())
        else:
            # internal resistance for discharging
            if self.__2D_RINT:
                resistance = self.__rint_dis_inter2d(temp, soc)
            else:
                resistance = self.__rint_dis_inter1d(soc)
            self.__log.debug('Resistance discharging stack: ' + str(resistance) + ' Ohm, resistance discharging module:'
                             + str(resistance / self.get_parallel_scale() * self.get_serial_scale()))
            return float(resistance / self.get_parallel_scale() * self.get_serial_scale())

    def get_cell_per_stack(self) -> int:
        return self.__CELL_NUMBER

    def get_min_voltage(self) -> float:
        return self.__MIN_CELL_VOLTAGE * self.get_cell_per_stack() * self.get_serial_scale()

    def get_max_voltage(self) -> float:
        return self.__MAX_CELL_VOLTAGE * self.get_cell_per_stack() * self.get_serial_scale()

    def get_self_discharge_current(self, redox_flow_state: RedoxFlowState) -> float:
        current_stack = redox_flow_state.current / self.get_parallel_scale()
        self_discharge_shunt = (float(self.__shunt_interp2d(redox_flow_state.soc_stack, current_stack)) *
                                self.get_cell_per_stack() * self.get_serial_scale() * self.get_parallel_scale())
        self_discharge_membrane = (self.__SELF_DISCHARGE_CURRENT_DENS_MEMBRANE * self.get_specific_cell_area() *
                                   self.get_cell_per_stack() * self.get_serial_scale() * self.get_parallel_scale())
        return self_discharge_membrane + self_discharge_shunt

    def get_specific_cell_area(self) -> float:
        return self.__CELL_AREA

    def get_electrode_thickness(self) -> float:
        return self.__ELECTRODE_THICKNESS / 100

    def get_electrode_porosity(self) -> float:
        return self.__ELECTRODE_POROSITY

    def get_hydraulic_resistance(self) -> float:
        return self.__HYDRAULIC_RESISTANCE / (self.get_serial_scale() * self.get_parallel_scale())

    def dependent_parameters(self) -> bool:
        return self.__DEPENDENT_PARAMETERS

    def close(self):
        super().close()
        self.__log.close()
        self.__electrolyte_system.close()



from configparser import ConfigParser

from simses.commons.config.simulation.simulation_config import SimulationConfig,  create_list_from, clean_split


class EnergyManagementConfig(SimulationConfig):
    """
    Energy management specific configs
    """

    SECTION: str = 'ENERGY_MANAGEMENT'

    STRATEGY: str = 'STRATEGY'
    POWER_FCR: str = 'POWER_FCR'
    POWER_IDM: str = 'POWER_IDM'
    SOC_SET: str = 'SOC_SET'
    MAX_POWER: str = 'MAX_POWER'
    MIN_SOC: str = 'MIN_SOC'
    MAX_SOC: str = 'MAX_SOC'
    FCR_RESERVE: str = 'FCR_RESERVE'
    MAX_POWER_MONTHLY: str = 'MAX_POWER_MONTHLY'
    MAX_POWER_MONTHLY_MODE: str = 'MAX_POWER_MONTHLY_MODE'
    EV_CHARGING_STRATEGY: str = 'EV_CHARGING_STRATEGY'

    MULTI_USE_STRATEGIES: str = 'MULTI_USE_STRATEGIES'
    ENERGY_ALLOCATION: str = 'ENERGY_ALLOCATION'
    POWER_ALLOCATION: str = 'POWER_ALLOCATION'
    RANKING: str = 'RANKING'
    OPTIMIZED_POWER_DIST_CASE: str = 'OPTIMIZED_POWER_DIST_CASE'
    OPTIMIZED_POWER_DIST_CASE_LIM: str = 'OPTIMIZED_POWER_DIST_CASE_LIM'

    def __init__(self, config: ConfigParser, path: str = None):
        super().__init__(path, config)

    @property
    def operation_strategy(self) -> str:
        """Returns operation strategy from __analysis_config file_name"""
        return self.get_property(self.SECTION, self.STRATEGY)

    @property
    def max_fcr_power(self) -> float:
        """Returns max power for providing frequency containment reserve from __analysis_config file_name"""
        return float(self.get_property(self.SECTION, self.POWER_FCR))

    @property
    def max_idm_power(self) -> float:
        """Returns max power for intra day market transactions from __analysis_config file_name"""
        return float(self.get_property(self.SECTION, self.POWER_IDM))

    @property
    def soc_set(self) -> float:
        """Returns the optimal soc for a FCR storage from __analysis_config file_name.
        In case of an overall efficiency below 1, the optimal soc should be higher than 0.5"""
        return float(self.get_property(self.SECTION, self.SOC_SET))

    @property
    def max_power(self) -> float:
        """Returns max power for peak shaving from __analysis_config file_name"""
        return float(self.get_property(self.SECTION, self.MAX_POWER))

    @property
    def min_soc(self) -> float:
        """Returns min soc from __analysis_config file_name"""
        return float(self.get_property(self.SECTION, self.MIN_SOC))

    @property
    def max_soc(self) -> float:
        """Returns max soc from __analysis_config file_name"""
        return float(self.get_property(self.SECTION, self.MAX_SOC))

    @property
    def fcr_reserve(self) -> float:
        """Returns max soc from __analysis_config file_name"""
        return float(self.get_property(self.SECTION, self.FCR_RESERVE))

    @property
    def max_power_monthly(self) -> [[str]]:
        """Returns a list of monthly max power """
        max_power_monthly = self.get_property(self.SECTION, self.MAX_POWER_MONTHLY)
        if max_power_monthly is None:
            max_power_monthly = [0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0]
        else:
            props: [str] = clean_split(max_power_monthly)
            max_power_monthly = create_list_from(props)
        return max_power_monthly

    @property
    def max_power_monthly_mode(self) -> bool:
        """Returns max power monthly from __analysis_config file_name"""
        try:
            get_mode = self.get_property(self.SECTION, self.MAX_POWER_MONTHLY_MODE)
            if get_mode == 'True':
                mode = True
            else:
                mode = False
        except (KeyError, TypeError):
            mode = False
        return mode

    @property
    def ev_charging_strategy(self) -> float:
        """Returns EV charging strategy from __analysis_config file_name"""
        return self.get_property(self.SECTION, self.EV_CHARGING_STRATEGY)

    @property
    def multi_use_strategies(self) -> [str]:
        """Returns multi-use strategies"""
        props: [str] = clean_split(self.get_property(self.SECTION, self.MULTI_USE_STRATEGIES), ',')
        return props

    @property
    def energy_allocation(self) -> [float]:
        """Returns energy allocation in multi-use scenario"""
        props: [float] = clean_split(self.get_property(self.SECTION, self.ENERGY_ALLOCATION), ',')
        return props

    @property
    def power_allocation(self) -> [float]:
        """Returns power allocation in multi-use scenario"""
        props: [float] = clean_split(self.get_property(self.SECTION, self.POWER_ALLOCATION), ',')
        return props

    @property
    def multi_use_rank(self) -> [int]:
        """Returns the different priorities of different strategies"""
        props: [int] = clean_split(self.get_property(self.SECTION, self.RANKING), ',')
        return props

    @property
    def optimized_power_dist_case(self) -> str:
        """Returns case for PSO Algorithm"""
        return self.get_property(self.SECTION, self.OPTIMIZED_POWER_DIST_CASE)

    @property
    def optimized_power_dist_case_lim(self) -> str:
        """Returns case for PSO Algorithm"""
        return self.get_property(self.SECTION, self.OPTIMIZED_POWER_DIST_CASE_LIM)


from simses.commons.config.simulation.general import GeneralSimulationConfig
from simses.commons.profile.power.power_profile import PowerProfile
from simses.commons.profile.technical.binary import BinaryProfile
from simses.commons.config.simulation.profile import ProfileConfig
from simses.commons.config.simulation.energy_management import EnergyManagementConfig
from simses.commons.state.energy_management import EnergyManagementState
from simses.commons.state.system import SystemState
from simses.logic.energy_management.strategy.operation_priority import OperationPriority
from simses.logic.energy_management.strategy.operation_strategy import OperationStrategy
import numpy as np

class ElectricVehicle(OperationStrategy):

    """
    ElectricVehicle is a basic operation strategy which simulates an EV during trips and recharge. The algorithm
    requires the following profiles and parameters:
    - A load profile that contains the required (and recuperated) power while driving
    - A binary profile that contains ones when the EV is parked & can be charged and zeros if not
    - A string for the parameter 'EV_CHARGING_STRATEGY' which can be:
        - "Uncontrolled": Always recharge, if SOC<100% and parked at "home"
        - "Mean_Power": This strategy checks the next departure time (perfect foresight) and chooses the charging power
            to reach 100% SOC exactly at departure time.
        - "Paused, integer threshold value": This strategy charges immediately after arrival until a threshold value
            (e.g. 0.8 -> 80% SOC). Then the charging is paused. The charging continues with maximal grid power at the
            time (t = Delta_E/P) that allows 100% SOC to be reached at the next departure time (with 30 minutes buffer).

    - The maximal grid power is set in the 'ENERGY_MANAGEMENT' section as 'MAX_POWER'
    - Using load and binary profile from emobpy, the LOAD_SCALING_FACTOR in [PROFILE] section should be set to 1

    Whenever the EV is not parked & ready to charge (binary=0), the EMS just follows the load profile.
    If the car is parked & ready to charge (binary=1), it is recharged depending on 'EV_CHARGING_STRATEGY'.
    """

    def __init__(self, config: GeneralSimulationConfig, profile_config: ProfileConfig, ems_config: EnergyManagementConfig,
                 power_profile: PowerProfile):

        super().__init__(OperationPriority.MEDIUM)

        self.__load_profile_driving: PowerProfile = power_profile
        self.__binary_profile = BinaryProfile(config, profile_config)

        self.__power: float = 0.0
        self.__binary: bool = False
        self.__soc_to_charge: float = 0.0
        self.__charging_strategy = ems_config.ev_charging_strategy
        self.__max_ac_grid_power = ems_config.max_power
        self.__fast_charging_event = 0
        self.__time_counter_postponed_trip = 0
        self.__postponed_trip_power = []
        self.__postponed_trip_binary = []
        self.__postponed_time = 0
        self.__Wh_to_Ws = 3600
        self.__next_time_zero = None
        self.__timestep = config.timestep

        self.__helper_class = HelperClassForProfileCopies(config, profile_config, ems_config)
        if self.__charging_strategy[0:6] == 'Paused':
            self.__threshold = float(self.__charging_strategy[-3:])

    def next(self, time: float, system_state: SystemState, power: float = 0) -> float:
        self.__power = self.__load_profile_driving.next(time)
        temp_binary = np.ceil(self.__binary_profile.next(time))
        if temp_binary!=0 and temp_binary!=1:
            raise ValueError("The binary profile contains values different from 0 or 1!")
        self.__binary = bool(temp_binary)

        if self.__binary != False and self.__binary != True:
            raise ValueError("The binary profile contains values different from 0 or 1!")

        if not self.__binary or self.__fast_charging_event > 0: # if EV is not at home: use load profile value
            # Possibly binary profile must also be shifted if fast charging is required on a long trip!
            # Right now, the break for fast charging is neglected (trip becomes shorter)
            if system_state.soc < 0.02:
                self.__fast_charging_event = 1  # if fast_charging has to start
                if self.__binary == 0:
                    self.__next_time_binary_one = self.__helper_class.binary_change_function(time, self.__binary)

            elif system_state.soc > 0.80 and self.__fast_charging_event == 1:
                self.__fast_charging_event = 2  # if fast_charging has happened and postponed period starts
                self.__postponed_time = len(self.__postponed_trip_power) * self.__timestep
                # Find next time when binary==1
                if self.__binary == 0:
                    self.__next_time_binary_one = self.__helper_class.binary_change_function(time, self.__binary)

            if self.__fast_charging_event == 0:
                power = self.__power    # normal case

            elif self.__fast_charging_event == 1:
                # 1st phase of fast-charging: charging with 150 kW (if possible by car) and save old power_values
                self.__postponed_trip_power.append(self.__power)
                self.__binary = True
                power = 150000

            elif self.__fast_charging_event == 2:
                # 2nd phase of fast-charging: apply old power_values and save current value from profile
                if time < self.__next_time_binary_one + self.__postponed_time:
                   power = self.__postponed_trip_power.pop(0)
                   self.__postponed_trip_power.append(self.__power)
                else:
                    self.__fast_charging_event = 0
                    power = self.__power
                    self.__postponed_trip_power = []
                self.__binary = False

        elif system_state.soc == 1:                 # if EV is at home and SOC=100%: do nothing
            power = 0
        elif self.__charging_strategy == 'Uncontrolled':   # if EV is at home and SOC!=100%: Recharge with charging strategy
            power = self.__max_ac_grid_power

        elif self.__charging_strategy == 'Mean_power':  # Charge with mean power to be fully charged at departure time

            if self.__next_time_zero is None:
                self.__next_time_zero = self.__helper_class.binary_change_function(time, self.__binary)

            charging_duration = self.__next_time_zero - time
            capacity_to_charge = (1 - system_state.soc) * system_state.capacity * self.__Wh_to_Ws
            if charging_duration > 0:
                power = capacity_to_charge / charging_duration
            else:
                power = self.__max_ac_grid_power
                self.__next_time_zero = None

            if charging_duration <= self.__timestep:
                self.__next_time_zero = None


        elif self.__charging_strategy[0:6] == 'Paused':

            if system_state.soc < self.__threshold:
                power = self.__max_ac_grid_power
            else:
                if self.__next_time_zero is None:
                    self.__next_time_zero = self.__helper_class.binary_change_function(time, self.__binary)
                else:
                    charging_duration = (1-system_state.soc)*system_state.capacity*self.__Wh_to_Ws/self.__max_ac_grid_power
                    charging_restart_time = self.__next_time_zero - charging_duration - 1800
                    # start recharging after the pause, buffer of 30 minutes (1800s)
                    if time < charging_restart_time:
                        power = 0
                    else:
                        power = self.__max_ac_grid_power
                        if self.__next_time_zero - time < self.__timestep:
                            self.__next_time_zero = None

        return power

    def update(self, energy_management_state: EnergyManagementState) -> None:
        energy_management_state.load_power = self.__power
        energy_management_state.binary = int(self.__binary)

    def clear(self) -> None:
        self.__power = 0.0

    def close(self) -> None:
        self.__binary_profile.close()
        self.__load_profile_driving.close()



class HelperClassForProfileCopies:

    def __init__(self, config: GeneralSimulationConfig, profile_config: ProfileConfig, ems_config: EnergyManagementConfig):

        # open second soc and binary profile for forecasts
        # This is necessary, as next function of profiles of instance of ElectricVehicleSOC-class saves previous values
        # Thus, those profiles cannot be used for forecasting SOC or binary values
        self.__binary_profile_copy = BinaryProfile(config, profile_config)
        self.__timestep = config.timestep

    def binary_change_function(self, time, current_binary):
        # This function returns the time when the binary profile changes its value
        # from 0 to 1 or 1 to 0

        return self.__binary_profile_copy.next_change_in_binary(time, current_binary)



import math

from simses.commons.state.technology.lithium_ion import LithiumIonState
from simses.technology.lithium_ion.cell.type import CellType
from simses.commons.log import Logger
from simses.technology.lithium_ion.degradation.calendar.calendar_degradation import \
    CalendarDegradationModel

# Source:
# Thomas Bank, Jan Feldmann, Sebastian Klamor, Stephan Bihn, Dirk Uwe Sauer:
# Extensive aging analysis of high-power lithium titanate oxide batteries:
# Impact of the passive electrode effect
# Journal of Power Sources 473 (2020) 228566, https://doi.org/10.1016/j.jpowsour.2020.228566


class LTONMCCalendarDegradationModel(CalendarDegradationModel):

    def __init__(self, cell_type: CellType):
        super().__init__(cell_type)
        self.__log: Logger = Logger(type(self).__name__)

        self.__capacity_loss = 0
        self.__initial_capacity = self._cell.get_nominal_capacity()
        self.__capacity_loss_cal = cell_type.get_calendar_capacity_loss_start()

        self.__A_QLOSS = -0.3517  # constant
        self.__B_QLOSS = 0.2275  # constant
        self.__C_QLOSS = -0.01189  # constant
        self.__D_QLOSS = 0.008428  # constant
        self.__E_QLOSS = -11.89  # constant
        self.__F_QLOSS = 107.7  # constant

        self.__resistance_increase = 0
        self.__rinc_calendar = 0

        self.__A_RINC = 1/35  # constant

    def calculate_degradation(self, time: float, battery_state: LithiumIonState) -> None:
        time_passed = time - battery_state.time
        soc = battery_state.soc
        # According to the source, if the SOC is smaller then 0.7 there is no calendar aging
        if soc < 0.70:
            capacity_loss = 0
        else:
            qloss: float = self.__capacity_loss_cal  # in pu
            virtual_time: float = ((1 - qloss) * 100 - 100) / (self.__A_QLOSS * soc + self.__B_QLOSS)
            if virtual_time > 35:
                virtual_time: float = ((1 - qloss) * 100 - (self.__E_QLOSS * soc + self.__F_QLOSS)) / (self.__C_QLOSS * soc + self.__D_QLOSS)
                rel_capacity_status = lambda time: ((self.__C_QLOSS * soc + self.__D_QLOSS) * time + (self.__E_QLOSS * soc + self.__F_QLOSS)) / 100
            else:
                rel_capacity_status = lambda time: ((self.__A_QLOSS * soc + self.__B_QLOSS) * time + 100) / 100
            total_time = virtual_time + time_passed / 86400  # in days
            capacity_loss = (1 - rel_capacity_status(total_time)) - self.__capacity_loss_cal
            # Total calendrical capacity loss
        self.__capacity_loss_cal += capacity_loss
        # Delta calendrical capacity loss
        self.__capacity_loss = capacity_loss * self.__initial_capacity

    def calculate_resistance_increase(self, time: float, battery_state: LithiumIonState) -> None:
        rinc = self.__rinc_calendar
        time_passed = time - battery_state.time
        virtual_time: float = (rinc * 100) / self.__A_RINC
        rel_resistance_increase = lambda time: (self.__A_RINC * time) / 100
        total_time = virtual_time + time_passed / 86400  # in days
        resistance_increase = (rel_resistance_increase(total_time)) - self.__rinc_calendar
        self.__rinc_calendar += resistance_increase
        self.__resistance_increase = resistance_increase

    def get_degradation(self) -> float:
        return self.__capacity_loss

    def get_resistance_increase(self) -> float:
        return self.__resistance_increase

    def reset(self, battery_state: LithiumIonState) -> None:
        self.__capacity_loss = 0
        self.__resistance_increase = 0

    def close(self) -> None:
        pass



import pytest

from simses.system.power_electronics.acdc_converter.notton import NottonAcDcConverter

charge = list((10.0, v, v) for v in range(0, 10, 1))
discharge = list((10.0, v, v) for v in range(-10, 0, 1))
charge_fail = list((10.0, v, 0.0) for v in range(0, 10, 1))
discharge_fail = list((10.0, v, 0.0) for v in range(-10, 0, 1))


@pytest.fixture(scope="function")
def uut(max_power: float):
    return NottonAcDcConverter(max_power)


@pytest.mark.parametrize('max_power, power, result', charge)
def test_to_dc(result, power, uut: NottonAcDcConverter):
    res = uut.to_dc(power, 0)
    print(res)
    assert result >= res >= result - 1.0


@pytest.mark.parametrize('max_power, power, result', discharge)
def test_to_ac(result, power, uut: NottonAcDcConverter):
    res = uut.to_ac(power, 0)
    print(res)
    assert result >= res >= result - 1.0


@pytest.mark.parametrize('max_power, power, result', charge)
def test_to_dc_reverse(result, power, uut: NottonAcDcConverter):
    res = uut.to_dc_reverse(power, 0)
    print(res)
    assert result + 1.0 >= res >= result


@pytest.mark.parametrize('max_power, power, result', discharge)
def test_to_ac_reverse(result, power, uut: NottonAcDcConverter):
    res = uut.to_ac_reverse(power, 0)
    print(res)
    assert result + 1.0 >= res >= result


@pytest.mark.parametrize('max_power, power, result', discharge_fail)
def test_to_dc_fail(result, power, uut: NottonAcDcConverter):
    res = uut.to_dc(power, 0)
    print(res)
    assert result >= res >= result - 1.0


@pytest.mark.parametrize('max_power, power, result', charge_fail)
def test_to_ac_fail(result, power, uut: NottonAcDcConverter):
    res = uut.to_ac(power, 0)
    print(res)
    assert result >= res >= result - 1.0


@pytest.mark.parametrize('max_power, power, result', discharge_fail)
def test_to_dc_reverse_fail(result, power, uut: NottonAcDcConverter):
    res = uut.to_dc_reverse(power, 0)
    print(res)
    assert result + 1.0 >= res >= result


@pytest.mark.parametrize('max_power, power, result', charge_fail)
def test_to_ac_reverse_fail(result, power, uut: NottonAcDcConverter):
    res = uut.to_ac_reverse(power, 0)
    print(res)
    assert result + 1.0 >= res >= result



import math
from abc import ABC, abstractmethod

from simses.commons.config.simulation.redox_flow import RedoxFlowConfig
from simses.commons.log import Logger
from simses.commons.state.technology.redox_flow import RedoxFlowState
from simses.technology.redox_flow.stack.electrolyte.abstract_electrolyte import ElectrolyteSystem


class StackModule(ABC):
    """A StackModule describes a module of electrical serial and parallel connected redox flow stacks."""

    def __init__(self, electrolyte_system: ElectrolyteSystem, voltage: float, power: float, cell_number_per_stack: int,
                 stack_power: float, redox_flow_config: RedoxFlowConfig):
        super().__init__()
        self.__log: Logger = Logger(self.__class__.__name__)
        self.__electrolyte_system: ElectrolyteSystem = electrolyte_system
        self.__exact_size = redox_flow_config.exact_size
        serial, parallel = self.__stack_connection(voltage, power, cell_number_per_stack, stack_power)
        self._SERIAL_SCALE: float = serial
        self._PARALLEL_SCALE: float = parallel
        self.__max_power = serial * parallel * stack_power * 2
        self.__log.debug('Stacks serial: ' + str(serial) + ', parallel: ' + str(parallel))

    def __stack_connection(self, voltage: float, power: float, cell_number_per_stack: int, stack_power: float):
        """
        Calculates the number of serial and parallel connected stacks in a stack module to obtain the system voltage and
        power.

        Parameters
        ----------
        voltage : float
            Voltage of the system in V.
        power : float
            Power of the system in W.
        cell_number_per_stack : int
            Number of cells per stack.
        stack_power : float
            Nominal power of one stack.

        Returns
        -------
            Serial number of stacks connected in a stack module.
            Parallel number of stacks connected in a stack module.
        """
        stack_voltage = (cell_number_per_stack * self.get_nominal_voltage_cell())
        if self.__exact_size:
            number_stacks: float = power / stack_power
            serial: float = min(voltage / stack_voltage, number_stacks)
            parallel: float = number_stacks / serial
            return serial, parallel
        else:
            # integer serial and parallel stack numbers, highest number used
            number_stacks: int = math.ceil(power / stack_power)
            serial: int = min(math.ceil(voltage / stack_voltage), number_stacks)
            parallel: int = math.ceil(number_stacks / serial)
            return serial, parallel

    @abstractmethod
    def get_open_circuit_voltage(self, redox_flow_state: RedoxFlowState) -> float:
        """
        Calculates the open circuit voltage (OCV) of a stack module depended on the electrolyte.

        Parameters
        ----------
        redox_flow_state : RedoxFlowState
            Current state of redox_flow.

        Returns
        -------
        float:
            OCV of the stack module in V.
        """
        pass

    @abstractmethod
    def get_nominal_voltage_cell(self) -> float:
        """
        Returns the nominal voltage of a single cell of a stack module in V. The value is used to change the capacity
        in Ws to its value in As and vice versa. The value of the OCV at SOC = 50 % and temperature = 30 °C is used.

        Returns
        -------
        float:
           Nominal cell voltage in V.
        """
        pass

    @abstractmethod
    def get_internal_resistance(self, redox_flow_state: RedoxFlowState) -> float:
        """
        Determines the internal resistance of the stack module based on the current RedoxFlowState.

        Parameters
        ----------
        redox_flow_state : RedoxFlowState
            Current state of redox_flow.

        Returns
        -------
        float:
            Internal resistance of the stack module in Ohm.
        """
        pass

    @abstractmethod
    def get_cell_per_stack(self) -> int:
        """
        Determines the cells per stack for the used stack type.

        Returns
        -------
        int:
            Number of cells per stack.
        """
        pass

    @abstractmethod
    def get_min_voltage(self) -> float:
        """
        Determines the minimal voltage of a stack module.

        Returns
        -------
        float:
            Minimal stack module voltage in V.
        """
        pass

    @abstractmethod
    def get_max_voltage(self) -> float:
        """
        Determines the maximal voltage of a stack module.

        Returns
        -------
        float:
            Maximal stack module voltage in V.
        """
        pass

    @abstractmethod
    def get_self_discharge_current(self, redox_flow_state: RedoxFlowState) -> float:
        """
        Determines the self-discharge current that discharges the stack module during standby and which accounts for
        self-discharge losses during cycling (Coulomb efficiency). The connection of multiple cells and stacks is
        considered.

        Parameters
        ----------
        redox_flow_state : RedoxFlowState
            Current state of redox_flow.

        Returns
        -------
        float:
            Total self-discharge current for a stack module in A.
        """
        pass

    @abstractmethod
    def get_specific_cell_area(self) -> float:
        """
        Returns the specific geometrical electrode area in cm^2. This corresponds to the area of the electrode that
        presses on the membrane.

        Returns
        -------
        float:
            Cell area in cm^2.
        """
        pass

    def get_electrode_thickness(self) -> float:
        """
        Returns the electrode thickness in m.

        Returns
        -------
        float:
            Electrode thickness in m.
        """
        pass

    @abstractmethod
    def get_electrode_porosity(self) -> float:
        """
        Returns the electrode porosity.

        Returns
        -------
        float:
            Electrode porosity.
        """
        pass

    @abstractmethod
    def get_hydraulic_resistance(self) -> float:
        """
        Returns the hydraulic resistance of the system in 1/m^3. The electrolyte flows through all stacks parallel.

        Returns
        -------
        float:
            Hydraulic resistance in 1/m^3.
        """
        pass

    def get_stacks_electrolyte_volume(self) -> float:
        """
        Returns the volume of electrolyte in the stack module electrodes in m^3 for one electrolyte side.

        Returns
        -------
        float:
            Electrolyte volume in the electrodes of the stack module in m^3.
        """
        stack_electrolyte_volume = (self.get_specific_cell_area() * self.get_electrode_thickness() / 10000 *
                                    self.get_cell_per_stack() * self.get_serial_scale() * self.get_parallel_scale() *
                                    self.get_electrode_porosity())
        if self.get_total_electrolyte_volume() < stack_electrolyte_volume:
            return self.get_total_electrolyte_volume()
        else:
            return stack_electrolyte_volume

    def get_total_electrolyte_volume(self) -> float:
        """
        Returns the total volume of the electrolyte of on electrolyte side (anolyte or catholyte).

        Returns
        -------
        float:
            Electrolyte volume of one side in m^3.
        """
        electrolyte_volume = (self.__electrolyte_system.get_start_capacity() * 3600 / (self.__electrolyte_system.FARADAY
                              * self.__electrolyte_system.get_vanadium_concentration() *
                              self.get_nominal_voltage_cell()))
        return electrolyte_volume

    def get_max_current_low_soc(self) -> float:
        """
        Determines the maximal current of the stack module at the lowest soc.

        Returns
        -------
        float:
            Maximal current in A.
        """
        return self.__max_power / self.get_min_voltage()

    def get_max_current_high_soc(self) -> float:
        """
        Determines the maximal current at the maximal voltage.

        Returns
        -------
        float:
            Maximal current in A at the maximal voltage.
        """
        return self.__max_power / 2 / self.get_max_voltage()

    def get_max_power(self) -> float:
        """
        Returns the maximal power of the redox flow system.

        Returns
        -------
        float:
            Maximal Power of the redox flow system in W.
        """
        return self.__max_power

    def dependent_parameters(self) -> bool:
        """
        Boolean value that signals whether parameters for the electrochemical model are dependent on voltage or current
        and therefore require iterations (e. q. the internal resistance is dependent on the current).

        Returns
        -------
        bool:
            True if parameters for the electrochemical model are dependent on voltage or current.
        """
        pass

    def get_name(self) -> str:
        """
        Determines the class name of a stack typ  (e. g. CellDataStack5500W).

        Returns
        -------
        str:
            Class name of a stack typ.
        """
        return self.__class__.__name__

    def get_serial_scale(self) -> float:
        """
        Returns the serial scale of stacks in the stack module. The value can be float if exact_size is true.

        Returns
        -------
        float:
            Number of serial stacks in the stack module.
        """
        return self._SERIAL_SCALE

    def get_parallel_scale(self) -> float:
        """
        Returns the parallel scale of stacks in the stack module. The value can be float if exact_size is true.

        Returns
        -------
        float:
            Number of parallel stacks in the stack module.
        """
        return self._PARALLEL_SCALE

    @abstractmethod
    def close(self):
        """Closing all resources in StackModule."""
        self.__log.close()
        self.__electrolyte_system.close()



import math

import numpy as np
import pandas as pd

from simses.commons.config.data.electrolyzer import ElectrolyzerDataConfig
from simses.commons.constants import Hydrogen
from simses.commons.state.technology.electrolyzer import ElectrolyzerState
from simses.technology.hydrogen.electrolyzer.stack.alkaline.fluid_model import \
    AlkalineFluidModel
from simses.technology.hydrogen.electrolyzer.stack.alkaline.pressure_model import \
    AlkalinePressureModel


class AlkalineElectricalModel:

    __POWER_HEADER = 'power in W'
    __TEMPERATURE_HEADER = 'temperature in °C'
    __PRESSURE_HEADER = 'P bar'
    __CURRENT_IDX = 3
    __TEMPERATURE_IDX = 0
    __PRESSURE_IDX = 1

    # Parameters TODO more general approach for other Electrolysers (here: HRI Stuart Electrolyser)
    __koh_weight_concentration = 30  # %
    __anode_surface = 0.03  # m^2
    __cathode_surface = 0.03  # m^2
    __anode_thickness = 0.002  # m
    __cathode_thickness = 0.002  # m
    __anode_membrane_gap = 0.00125  # m
    __cathode_membrane_gap = 0.00125  # m
    __membrane_surface_area = 0.03  # m^2
    __limiting_current_density = 300000  # A/m^2 (limiting current density)
    __standard_temperature = 25  # °C (Room Temperature)
    __standard_temperature = __standard_temperature + 273.15  # K

    def __init__(self, pressure_model: AlkalinePressureModel, fluid_model: AlkalineFluidModel, electrolyzer_data_config: ElectrolyzerDataConfig, parameters):
        self.__pressure_model = pressure_model
        self.__fluid_model = fluid_model
        self.__POWER_FILE = electrolyzer_data_config.alkaline_electrolyzer_multidim_lookup_currentdensity_file
        self.__power_curve = pd.read_csv(self.__POWER_FILE, compression='gzip')
        self.__parameters = parameters

    def get_cell_voltage(self, current: float, state: ElectrolyzerState):
        # Resistance correction factors
        r1 = self.__parameters.r1
        r2 = self.__parameters.r2
        r3 = self.__parameters.r3

        stack_temperature = state.temperature  # °C
        stack_temperature = stack_temperature + 273.15  # K

        v_cell = 0

        # Get theoretical cell voltage
        v_theoretical = self.__theoretical_voltage(state, current)
        v_cell += v_theoretical
        # Get activation overvoltages
        v_activation_anode = self.__activation_voltage_anode(state, current)
        v_activation_cathode = self.__activation_voltage_cathode(state, current)
        v_activation_total = v_activation_anode + v_activation_cathode
        v_cell += v_activation_total
        # Calculate ohmic overvoltages
        r_electrodes = self.__electrode_resistance()
        v_electrode_resistances = r_electrodes * current
        r_electrolyte = self.__electrolyte_resistance(current, stack_temperature)
        v_electrolyte_resistance = r_electrolyte * current
        r_membrane = self.__membrane_resistance(stack_temperature)
        v_membrane_resistance = r_membrane * current

        v_cell += v_electrolyte_resistance * r1
        v_cell += v_electrode_resistances * r2
        v_cell += v_membrane_resistance * r3

        return v_cell

    def get_current(self, power_cell: float, state: ElectrolyzerState):
        pressure = state.pressure_cathode     # bar (pressure assumed to be identical for cathode and anode) TODO Check Validity
        stack_temperature = state.temperature     # °C
        # Get current from lookuptable
        if power_cell <= 0:
            return 0
        else:
            temperature_idx = abs(self.__power_curve[self.__TEMPERATURE_HEADER] - stack_temperature).idxmin()
            filter_temperature = self.__power_curve.iloc[temperature_idx, self.__TEMPERATURE_IDX]   # °C
            powercurve = self.__power_curve[self.__power_curve[self.__TEMPERATURE_HEADER] == filter_temperature]
            powercurve.reset_index(drop=True, inplace=True)

            pressure_idx = abs(powercurve[self.__PRESSURE_HEADER] - pressure).idxmin()
            filter_pressure = powercurve.iloc[pressure_idx, self.__PRESSURE_IDX]  # Bar
            powercurve = powercurve[powercurve[self.__PRESSURE_HEADER] == filter_pressure]
            powercurve.reset_index(drop=True, inplace=True)

            power_idx = abs(powercurve[self.__POWER_HEADER] - power_cell).idxmin()
            current = powercurve.iloc[power_idx, 3]  # A
            return current


    def __reversible_voltage(self, stack_temperature) -> float:
        # Thermodynamic approximation of the reversible voltage at standard conditions
        return 1.50342 - 9.956 * 10 ** (-4) * stack_temperature + 2.5 * 10 ** (-7) * stack_temperature ** 2

    def __theoretical_voltage(self, state: ElectrolyzerState, current) -> float:
        stack_temperature = state.temperature           # °C
        stack_temperature = stack_temperature + 273.15  # K
        pressure = state.pressure_cathode  # bar (pressure assumed to be identical for cathode and anode) TODO Check Validity

        v_reversible_standard_conditions = self.get_rev_voltage(stack_temperature)

        molarity = self.__fluid_model.get_molarity(stack_temperature)
        # Calculate vapour pressure of pure water pw* and pressure of gaseous KOH solution pw
        # (From: "The Thermodynamics of Aqueous Water Electrolysis" by LeRoy)
        vapour_pressure_pure_water = self.__pressure_model.get_vapour_pressure_pure_water(stack_temperature)
        aqueous_vapour_pressure_koh = self.__pressure_model.get_aqueous_vapour_pressure_koh(molarity, stack_temperature)

        # Calculate water electrolysis theoretical voltage
        v_theoretical = v_reversible_standard_conditions + (
                (Hydrogen.IDEAL_GAS_CONST * stack_temperature) / (2 * Hydrogen.FARADAY_CONST)) * math.log(
            (pressure - aqueous_vapour_pressure_koh) ** (3 / 2) * (vapour_pressure_pure_water / aqueous_vapour_pressure_koh))

        # Correction terms for the consideration of real gas instead of ideal gas
        v_correction1 = (pressure - aqueous_vapour_pressure_koh) * (21.661 * 10 ** (-6) - (5.471 * 10 ** (-3)) / stack_temperature)
        v_theoretical += v_correction1
        v_correction2 = (pressure - aqueous_vapour_pressure_koh) ** 2 * (
                -1 * ((6.289 * 10 ** -6) / stack_temperature) + ((0.135 * 10 ** -3) / stack_temperature ** 1.5)
                + ((2.547 * 10 ** -3) / stack_temperature ** 2) - (0.4825 / stack_temperature ** 3))
        v_theoretical += v_correction2

        return v_theoretical

    def __activation_voltage_anode(self, state: ElectrolyzerState, current: float) -> float:
        # Anode correction factors
        aa = self.__parameters.aa
        ba = self.__parameters.ba
        ca = self.__parameters.ca

        # Transfer coefficient correction factor
        a1 = self.__parameters.a1

        stack_temperature = state.temperature  # °C
        stack_temperature = stack_temperature + 273.15  # K

        # Calculate bubble rate coverage of electrodes (assumed to be identical for both electrodes)
        bubble_rate_coverage = self.__bubble_rate_coverage(current, stack_temperature)
        # Calculate nominal electrode surfaces
        anode_surface_nom = self.__anode_surface * (1 - bubble_rate_coverage)
        current_density_anode = current / (anode_surface_nom * 10000.0)  # A/cm2
        # Calculate transfer coefficient for anode for HRI Electrolyser
        alpha_a = 0.0675 + 0.00095 * stack_temperature * a1
        # Calculate tafel slope for anode
        factor_anode = (Hydrogen.IDEAL_GAS_CONST * stack_temperature) / (
                2 * Hydrogen.FARADAY_CONST * alpha_a)
        # Calculate exchange current density for anode for HRI Electrolyser (Ni Electrode)
        exchange_current_density_anode = aa + ba * stack_temperature + ca * stack_temperature ** 2  # mA/cm2
        # Calculate activation overvoltage of the electrode:
        v_activation_anode = factor_anode * np.arcsinh(
            current_density_anode / (2 * exchange_current_density_anode / 1000.0))

        return v_activation_anode

    def __activation_voltage_cathode(self, state: ElectrolyzerState, current: float) -> float:
        # Cathode correction factors
        ac = self.__parameters.ac
        bc = self.__parameters.bc
        cc = self.__parameters.cc

        # Transfer coefficient correction factor
        a2 = self.__parameters.a2

        stack_temperature = state.temperature  # °C
        stack_temperature = stack_temperature + 273.15  # K

        # Calculate bubble rate coverage of electrodes (assumed to be identical for both electrodes)
        bubble_rate_coverage = self.__bubble_rate_coverage(current, stack_temperature)
        # Calculate nominal electrode surface
        cathode_surface_nom = self.__cathode_surface * (1 - bubble_rate_coverage)
        current_density_cathode = current / (cathode_surface_nom * 10000.0)
        # Calculate transfer coefficient for cathode for HRI Electrolyser
        alpha_c = 0.1175 + 0.00095 * stack_temperature * a2
        # Calculate tafel slope for cathode
        factor_cathode = (Hydrogen.IDEAL_GAS_CONST * stack_temperature) / (
                2 * Hydrogen.FARADAY_CONST * alpha_c)
        # Calculate exchange current density for cathode for HRI Electrolyser (Ni Electrode)
        exchange_current_density_cathode = ac + bc * stack_temperature + cc * stack_temperature ** 2  # mA/cm2
        # Calculate activation overvoltage of the electrode:
        v_activation_cathode = factor_cathode * np.arcsinh(
            current_density_cathode / (2 * exchange_current_density_cathode / 1000.0))

        return v_activation_cathode

    def __bubble_rate_coverage(self, current, stack_temperature) -> float:
        bubble_rate_coverage = (-97.25 + 182 * (stack_temperature / self.__standard_temperature) - 84 * (stack_temperature / self.__standard_temperature) ** 2) * (
                (current / self.__cathode_surface) / self.__limiting_current_density) ** 0.3

        return bubble_rate_coverage

    def get_rev_voltage(self, stack_temperature) -> float:
        """ Returns reversible Cell Voltage at standard pressure depending on stack temperature in K """
        return self.__reversible_voltage(stack_temperature)

    def calculate_bubble_rate_coverage(self, current, stack_temperature) -> float:
        return self.__bubble_rate_coverage(current, stack_temperature)

    def __electrode_resistance(self) -> float:
        # Calculate electrical conductivity of Ni TODO more general approach?
        conductivity_nickel = 14.3 * 10 ** 6  # S/m (generic value for nickel based electrode)
        conductivity_nickel = conductivity_nickel / 100  # S/cm
        # Calculate electrode resistances
        resistance_anode = (1 / conductivity_nickel) * ((self.__anode_thickness * 100) / (self.__anode_surface * 10000))
        resistance_cathode = (1 / conductivity_nickel) * ((self.__cathode_thickness * 100) / (self.__cathode_surface * 10000))

        return resistance_anode + resistance_cathode

    def __electrolyte_resistance(self, current, stack_temperature) -> float:
        molarity = self.__fluid_model.get_molarity(stack_temperature)
        bubble_rate_coverage = self.calculate_bubble_rate_coverage(current, stack_temperature)

        # Calculate ionic conductivity of KOH in S/cm
        conductivity_koh_free = -2.041 * molarity - 0.0028 * molarity ** 2 + 0.005332 * molarity * stack_temperature + 207.2 * (
                molarity / stack_temperature) \
                         + 0.001043 * molarity ** 3 - 0.0000003 * molarity ** 2 * stack_temperature ** 2
        # Calculate free resistance of electrolyte
        resistance_ele_free = (1 / conductivity_koh_free) * (
                    (self.__anode_membrane_gap / self.__anode_surface + self.__cathode_membrane_gap / self.__cathode_surface) / 100)
        # Incorporate bubble rate into resistance
        resistance_ele_epsilon = resistance_ele_free * ((1 / ((1 - (2 / 3) * bubble_rate_coverage) ** (3 / 2))) - 1)
        resistance_ele = resistance_ele_free + resistance_ele_epsilon

        return resistance_ele

    def __membrane_resistance(self, stack_temperature) -> float:
        # Generic values from "VERMEIREN, P. (1998). Evaluation of the ZirfonS separator for use in alkaline water electrolysis and Ni-H2 batteries.
        # International Journal of Hydrogen Energy, 23(5), 321–324. doi:10.1016/s0360-3199(97)00069-4" (FROM GRAPH!)
        resistance_membrane_per_area = self.__parameters.zirfon1 * stack_temperature ** 2 + self.__parameters.zirfon2 * stack_temperature + self.__parameters.zirfon3
        resistance_membrane_total = resistance_membrane_per_area / (self.__membrane_surface_area * 10000)

        return resistance_membrane_total

    def get_geometric_area_cell(self) -> float:
        return self.__membrane_surface_area




import os
import pathlib
from configparser import ConfigParser
from simses.commons.config.analysis.analysis_config import AnalysisConfig
from simses.commons.utils.utilities import convert_path_codec


class GeneralAnalysisConfig(AnalysisConfig):

    """
    General analysis configs
    """
    SECTION: str = 'GENERAL'
    SIMULATION: str = 'SIMULATION'
    SYSTEM_ANALYSIS: str = 'SYSTEM_ANALYSIS'
    LITHIUM_ION_ANALYSIS: str = 'LITHIUM_ION_ANALYSIS'
    REDOX_FLOW_ANALYSIS: str = 'REDOX_FLOW_ANALYSIS'
    HYDROGEN_ANALYSIS: str = 'HYDROGEN_ANALYSIS'
    SITE_LEVEL_ANALYSIS: str = 'SITE_LEVEL_ANALYSIS'
    PLOTTING: str = 'PLOTTING'
    TECHNICAL_ANALYSIS: str = 'TECHNICAL_ANALYSIS'
    ECONOMICAL_ANALYSIS: str = 'ECONOMICAL_ANALYSIS'
    EXPORT_ANALYSIS_TO_CSV: str = 'EXPORT_ANALYSIS_TO_CSV'
    PRINT_RESULTS_TO_CONSOLE: str = 'PRINT_RESULTS_TO_CONSOLE'
    EXPORT_ANALYSIS_TO_BATCH: str = 'EXPORT_ANALYSIS_TO_BATCH'
    MERGE_ANALYSIS: str = 'MERGE_ANALYSIS'

    def __init__(self, config: ConfigParser, path: str = None):
        super().__init__(path, config)

    def get_result_for(self, path: str) -> str:
        """Returns name of the simulation to analyse."""
        simulation = convert_path_codec(self.get_property(self.SECTION, self.SIMULATION))
        if simulation == 'LATEST':
            result_dirs = list()
            tmp_dirs = os.listdir(path)
            # res = filter(os.path.isdir, tmp_dirs)
            for dir in tmp_dirs:
                if os.path.isdir(path + dir):
                    result_dirs.append(path + dir + '/')
            return max(result_dirs, key=os.path.getmtime)
        elif os.path.isabs(simulation):
            return pathlib.Path(simulation).as_posix() + '/'
        else:
            return path + simulation + '/'

    @property
    def system_analysis(self) -> bool:
        """Returns boolean value for system_analysis after the simulation"""
        return self.get_property(self.SECTION, self.SYSTEM_ANALYSIS) in ['True']

    @property
    def lithium_ion_analysis(self) -> bool:
        """Returns boolean value for lithium_ion_analysis after the simulation"""
        return self.get_property(self.SECTION, self.LITHIUM_ION_ANALYSIS) in ['True']

    @property
    def redox_flow_analysis(self) -> bool:
        """Returns boolean value for redox_flow_analysis after the simulation"""
        return self.get_property(self.SECTION, self.REDOX_FLOW_ANALYSIS) in ['True']

    @property
    def hydrogen_analysis(self) -> bool:
        """Returns boolean value for redox_flow_analysis after the simulation"""
        return self.get_property(self.SECTION, self.HYDROGEN_ANALYSIS) in ['True']

    @property
    def site_level_analysis(self) -> bool:
        """Returns boolean value for redox_flow_analysis after the simulation"""
        return self.get_property(self.SECTION, self.SITE_LEVEL_ANALYSIS) in ['True']

    @property
    def plotting(self) -> bool:
        """Returns boolean value for matplot_plotting after the simulation"""
        return self.get_property(self.SECTION, self.PLOTTING) in ['True']

    @property
    def technical_analysis(self) -> bool:
        """Returns boolean value for technical analysis directly after the simulation"""
        return self.get_property(self.SECTION, self.TECHNICAL_ANALYSIS) in ['True']

    @property
    def economical_analysis(self) -> bool:
        """Returns boolean value for economical analysis directly after the simulation"""
        return self.get_property(self.SECTION, self.ECONOMICAL_ANALYSIS) in ['True']

    @property
    def export_analysis_to_csv(self) -> bool:
        """Defines if analysis results are to be exported to csv files"""
        return self.get_property(self.SECTION, self.EXPORT_ANALYSIS_TO_CSV) in ['True']

    @property
    def print_result_to_console(self) -> bool:
        """Defines if analysis results are to be printed to console"""
        return self.get_property(self.SECTION, self.PRINT_RESULTS_TO_CONSOLE) in ['True']

    @property
    def export_analysis_to_batch(self) -> bool:
        """Defines if analysis results are written to batch files"""
        return self.get_property(self.SECTION, self.EXPORT_ANALYSIS_TO_BATCH) in ['True']

    @property
    def merge_analysis(self) -> bool:
        """Defines if analysis results are merged"""
        return self.get_property(self.SECTION, self.MERGE_ANALYSIS) in ['True']

    @property
    def logo_file(self) -> str:
        return self.get_data_path(self.get_property(self.SECTION, 'LOGO_FILE'))



from simses.system.dc_coupling.dc_coupler import DcCoupling
from simses.system.dc_coupling.generation.pv_dc_generation import PvDcGeneration
from simses.system.dc_coupling.load.dc_bus_charging_random import DcBusChargingRandom


class BusChargingDcCoupling(DcCoupling):

    def __init__(self, charging_power: float, generation_power: float):
        super().__init__(DcBusChargingRandom(charging_power), PvDcGeneration(generation_power))



import os
import pathlib
from abc import ABC
from configparser import ConfigParser

from simses.commons.utils.utilities import convert_path_codec
from simses.data import DATA_DIR


class Config(ABC):
    """
    The Config class contains all necessary configuration options of a package, e.g. simulation or analysis. In addition,
    the Config class selects proper options from defaults, local or in code configurations. For all sections of each
    config a own config class is provided and is inherited from Config.

    Configs are taken from INI files in config package. *.defaults.ini are read in first, followed by *.local.ini and
    finally overwritten by a ConfigParser passed in as a constructor argument. This threefold delivers functionality by
    default if you just want to run a possible setup. If a specific simulation should be run locally, the config parameters
    of the *.local.ini file will overwrite the configuration of defaults; only necessary parameters can be included in
    the local config file. Furthermore, for sensitivity analysis the ConfigParser argument is taken into account to
    automate various configurations. The ConfigParser should have the same structure as the config files and overwrites
    the given values.
    """

    CONFIG_EXT: str = '.ini'
    DEFAULTS: str = '.defaults' + CONFIG_EXT
    LOCAL: str = '.local' + CONFIG_EXT

    __used_config: dict = dict()
    __extensions: [str] = [DEFAULTS, LOCAL, CONFIG_EXT]

    def __init__(self, path: str, name: str, config: ConfigParser):
        """
        Constructor of Config

        Parameters
        ----------
        path :
            path to config file folder
        name :
            name of config file
        config :
            ConfigParser overwriting values from config file
        """
        # if path is None:
        #     path = get_path_for(name + self.DEFAULTS)
        self.__config: ConfigParser = ConfigParser()
        for extension in self.__extensions:
            self.__config.read(path + name + extension)
        self.__file_name: str = name + self.CONFIG_EXT
        self.__overwrite_config_with(config)

    def get_property(self, section: str, option: str):
        """
        Returns the value for given section and option

        Parameters
        ----------
        section :
            section of config
        option :
            option of config

        Returns
        -------

        """
        value = None
        try:
            value = self.__config[section][option]
            self.__add_to_used_config(section, option, value)
        except KeyError as err:
            raise err
        finally:
            return value

    def __overwrite_config_with(self, config: ConfigParser):
        if config is not None:
            for section in config.sections():
                if section in self.__config.sections():
                    for option in config.options(section):
                        if option in self.__config.options(section):
                            value = config[section][option]
                            # print('[' + type(self).__name__ + '] Setting new value in section ' + section +
                            #       ' for option ' + option + ' with ' + value)
                            self.__config[section][option] = value
                        else:
                            # print('Option ' + option + ' is added to section ' + section)
                            value = config[section][option]
                            self.__config[section][option] = value

    def __add_to_used_config(self, section: str, option: str, value: str):
        key: str = self.__file_name
        if key not in self.__used_config.keys():
            self.__used_config[key] = ConfigParser()
        config: ConfigParser = self.__used_config[key]
        if section not in config.sections():
            config.add_section(section)
        if option not in config.options(section):
            config.set(section, option, value)

    def write_config_to(self, path: str) -> None:
        """
        Write current config to a file in given path

        Parameters
        ----------
        path :
            directory in which config file should be written

        Returns
        -------

        """
        # TODO how to write only used configs?
        # used_config: ConfigParser = self.__used_config[self.__file_name]
        with open(path + self.__file_name, 'w') as configfile:
            self.__config.write(configfile)
            # used_config.write(configfile)

    def get_data_path(self, path) -> str:
        path = convert_path_codec(path)
        if os.path.isabs(path):
            return pathlib.Path(path).as_posix() + '/'
        return DATA_DIR + path



from simses.commons.config.simulation.electrolyzer import ElectrolyzerConfig
from simses.commons.constants import Hydrogen
from simses.commons.state.technology.electrolyzer import ElectrolyzerState
from simses.system.auxiliary.pump.abstract_pump import Pump
from simses.system.auxiliary.water_heating.water_heater import WaterHeating
from simses.technology.hydrogen.control.thermal.ideal_var_flow import \
    IdealVarFlowThermalController
from simses.technology.hydrogen.electrolyzer.stack.stack_model import ElectrolyzerStackModel
from simses.technology.hydrogen.electrolyzer.thermal.abstract_model import ThermalModel


class SimpleThermalModel(ThermalModel):
    """ This model functions at Electrolyzer Level.
    This model calculates the temperaturechange in the electrlyzer stack
    the elelctrolyzer is represented by a area element"""

    def __init__(self, electrolyzer: ElectrolyzerStackModel, water_heating: WaterHeating, pump: Pump,
                 temperature: float, config: ElectrolyzerConfig):
        super().__init__()
        self.__HEAT_CAPACITY = electrolyzer.get_heat_capacity_stack() # J/K
        nominal_stack_power = electrolyzer.get_nominal_stack_power()  # W
        self.__max_water_flow_stack = self.__calculate_max_water_flow_stack(nominal_stack_power)  # mol/s
        self.__min_water_flow_stack = 0.1 * self.__max_water_flow_stack  # mol/s
        self.__thermal_controller = IdealVarFlowThermalController(config, self.__HEAT_CAPACITY)
        self.__temperature_stack_1 = 0  # °C
        self.__NUMBER_CELLS = electrolyzer.get_number_cells()
        self.__GEOM_AREA_STACK = electrolyzer.get_geom_area_stack() # cm²
        self.__GEOM_AREA_CELL = self.__GEOM_AREA_STACK / self.__NUMBER_CELLS  # cm²
        self.__convection_heat = 0  # initialize variable
        self.__heat_generation = 0  # initialize variable
        self.__heat_vape = 0  # initialize variable
        self.__TH_NEUT_VOLTAGE = 286 * 10 ** 3 / (2 * Hydrogen.FARADAY_CONST)  # V
        self.__SURFACE_RATIO_FRAME = self.__calculate_surface_ratio_frame(nominal_stack_power)
        self.__SURFACE_RATIO_ENDPLATE = self.__calculate_surface_ratio_endplates(nominal_stack_power)
        self.__SURFACE_RADTIO_TUBES = self.__calculate_surface_ratio_tubes(nominal_stack_power)
        self.__SURFACE_RATIO_SEPARATOR = self.__calculate_surface_ratio_separator(nominal_stack_power)
        self.__HEAT_TRANSMISSION_COEF_TUBES = self.__calculate_heat_transmission_coef_tubes(nominal_stack_power)
        self.__AMOUNT_WATER = electrolyzer.get_water_in_stack() # kg
        self.__h2o_flow_stack = 0  # mol/s
        self.__heat_h2o = 0  #
        self.__pump = pump
        self.__pump_power = 0  # W
        self.__tube_pressure_loss = 0  # bar
        self.__water_heating = water_heating
        self.__ambient_temperature: float = temperature - 273.15  # C

    def calculate(self, time: float, state: ElectrolyzerState, pressure_cathode_0, pressure_anode_0) -> None:
        timestep = time - state.time
        temperature_stack_0 = state.temperature  # °C
        current_dens = state.current / self.__GEOM_AREA_CELL # A/cm²
        sat_p_h2o = state.sat_pressure_h2o  # bar
        h2_genearation_per_area = state.hydrogen_production / self.__GEOM_AREA_STACK  # mol/s/cm²
        o2_generation_per_area = state.oxygen_production / self.__GEOM_AREA_STACK  # mol/s/cm²
        h2o_use_per_area = state.water_use / self.__GEOM_AREA_STACK  # mol/s/cm²

        # water gives all its heatenergy to the stack
        temperature_h2o_out = temperature_stack_0  # °C

        # specific heat generation in electrolyseur cell
        spec_heat_cell = self.__calculate_spec_heat_cell(state.voltage, current_dens, temperature_stack_0,
                                                         h2_genearation_per_area, o2_generation_per_area,
                                                         pressure_cathode_0, pressure_anode_0, sat_p_h2o)

        # specific heat dissipation through frame, endplates, tubes and gas separator
        spec_heat_frame = self.__calculate_spec_heat_frame(temperature_stack_0, self.__ambient_temperature) # W/cm²
        spec_heat_endplates = self.__calculate_spec_heat_endplate(temperature_stack_0, self.__ambient_temperature)  # W/cm²
        spec_heat_tubes = self.__calculate_spec_heat_tubes(temperature_stack_0, self.__ambient_temperature)  # W/cm 2
        spec_heat_separator = self.__calculate_spec_heat_separator(temperature_stack_0, self.__ambient_temperature)

        # specific heat sink because of freshwater
        spec_heat_fresh_water = self.__calculate_spec_heat_freshwater(h2o_use_per_area, temperature_h2o_out,
                                                                      self.__ambient_temperature)

        # total heat generation electrolyzer stack
        heat_stack = (spec_heat_cell - spec_heat_frame - 2 * spec_heat_endplates - spec_heat_tubes - spec_heat_separator
                      - spec_heat_fresh_water) * self.__GEOM_AREA_STACK # W

        # calculation of watertemperature and waterflow for tempering the stack
        self.__thermal_controller.calculate(temperature_stack_0, heat_stack, timestep, current_dens)
        temperature_h2o_in = temperature_stack_0 + self.__thermal_controller.get_delta_water_temp_in()  # °C
        self.__h2o_flow_stack = self.__thermal_controller.get_h2o_flow()  # mol/s
        self.__heat_h2o = self.__calculate_heat_h2o(self.__h2o_flow_stack, temperature_h2o_in, temperature_h2o_out)

        # self.__temperature_stack_1 = temperature_stack_0 + timestep / self.__HEAT_CAPACITY * (
        #             heat_stack + self.__heat_h2o) + 273.15  # K -> °C
        # temperature calculation stack
        if self.__thermal_controller.get_heat_control_on():
            self.__temperature_stack_1 = temperature_stack_0 + timestep / self.__HEAT_CAPACITY * (heat_stack + self.__heat_h2o)  # °C
            # if heat control is on, water is circulated through the stack and controls the temperature of the stack
            # with its own temperature-> heat capacity of water is not accounted
        else:
            self.__temperature_stack_1 = temperature_stack_0 + timestep / (self.__HEAT_CAPACITY + self.__AMOUNT_WATER *
                                                                           Hydrogen.HEAT_CAPACITY_WATER) * \
                                         (heat_stack + self.__heat_h2o)  # °C
            # if heat control is off there is no water circulation -> heat capacity of water which remains in the stack
            # needs to be accounted for the cooling process

        # convection heat: heat that is transported to the ambient area
        if self.__heat_h2o < 0:
            self.__convection_heat = (spec_heat_frame + spec_heat_endplates + spec_heat_tubes + spec_heat_separator) * \
                                     self.__GEOM_AREA_STACK * self.__NUMBER_CELLS - self.__heat_h2o
        else:
            self.__convection_heat = (spec_heat_frame + spec_heat_endplates + spec_heat_tubes + spec_heat_separator) * \
                                     self.__GEOM_AREA_STACK * self.__NUMBER_CELLS

    def __calculate_spec_heat_freshwater(self, h2o_use_per_area, temp_h2o_in, temp_ambient) -> float:
        """ calculates the cooling effect of the feedwater
        if temp_h2o_in > temp_ambient: feedwater cools the water in the circulation"""
        delta_temp = temp_ambient - temp_h2o_in
        return Hydrogen.HEAT_CAPACITY_WATER * h2o_use_per_area * Hydrogen.MOLAR_MASS_WATER * delta_temp

    def __calculate_spec_heat_frame(self, temp_stack, temp_abient) -> float:
        delta_temp = temp_stack - temp_abient
        k_frame = 2.5  # W/(m² K)
        return k_frame * self.__SURFACE_RATIO_FRAME * delta_temp  # W/cm²

    def __calculate_spec_heat_endplate(self, temp_stack, temp_ambient) -> float:
        delta_temp = temp_stack - temp_ambient
        k_endplate = 3.6  # # W/(m² K)
        return k_endplate * self.__SURFACE_RATIO_ENDPLATE * delta_temp  # W/cm²

    def __calculate_spec_heat_tubes(self, temp_stack, temp_ambient) -> float:
        delta_temp = temp_stack - temp_ambient
        return self.__HEAT_TRANSMISSION_COEF_TUBES * self.__SURFACE_RADTIO_TUBES * delta_temp  # W/cm²

    def __calculate_spec_heat_separator(self, temp_stack, temp_ambient) -> float:
        delta_temp = temp_stack - temp_ambient
        k_separator = 9  # W/(m² K)
        return k_separator * self.__SURFACE_RATIO_SEPARATOR * delta_temp

    def __calculate_heat_h2o(self, h2o_flow_stack, temp_h2o_in, temp_h2o_out) -> float:
        delta_temp = temp_h2o_in - temp_h2o_out
        self.__water_heating.calculate_heating_power(h2o_flow_stack, delta_temp)
        return self.__water_heating.get_heating_power()

    def __calculate_spec_heat_generation_cell(self, cell_voltage, current_dens) -> float:
        return (cell_voltage / self.__NUMBER_CELLS - self.__TH_NEUT_VOLTAGE) * current_dens

    def __calcutlate_spec_heat_vape_h2o(self, temp_stack, h2_gen_per_area, o2_gen_per_area, p_c_0, p_a_0,
                                        sat_p_h2o) -> float:
        if h2_gen_per_area > 0:
            h2o_steam_flow_per_area = (sat_p_h2o / (1 + p_c_0 - sat_p_h2o)) * h2_gen_per_area + (sat_p_h2o /
                                                (1 + p_a_0 - sat_p_h2o)) * o2_gen_per_area  # mol/s/cm²
        else:
            h2o_steam_flow_per_area = 0
        return - (-43.79 * temp_stack + 4.509 * 10 ** 4) * h2o_steam_flow_per_area  # W/cm²

    def __calculate_spec_heat_cell(self, cell_voltage, current_dens, temp_stack, h2_gen_per_area, o2_gen_per_area,
                                   p_c_0, p_a_0, sat_p_h2o) -> float:
        spec_heat_vape_h2o = self.__calcutlate_spec_heat_vape_h2o(temp_stack, h2_gen_per_area, o2_gen_per_area, p_c_0,
                                                                  p_a_0, sat_p_h2o)
        spec_heat_gen_cell = self.__calculate_spec_heat_generation_cell(cell_voltage, current_dens)
        return spec_heat_gen_cell + spec_heat_vape_h2o

    def __calculate_surface_ratio_frame(self, nominal_stack_power) -> float:
        m_frame = -591 / 5750  # 1/kW
        c_frame = 357.1
        return (m_frame * nominal_stack_power / 1000 + c_frame) * 10 ** (-7)  # m²/cm²

    def __calculate_surface_ratio_endplates(self, nominal_stack_power) -> float:
        m_endplate = -129 / 11500  # 1/kW
        c_endplate = 19
        return (m_endplate * nominal_stack_power / 1000 + c_endplate) * 10 ** (-7)  # m²/cm²

    def __calculate_surface_ratio_tubes(self, nominal_stack_power) -> float:
        m_tubes = -107 / 115000  # 1/kW
        c_tubes = 2.06
        return (m_tubes * nominal_stack_power / 1000 + c_tubes) * 10 ** (-5)  # m²/cm²

    def __calculate_heat_transmission_coef_tubes(self, nominal_stack_power) -> float:
        m_tubes = 53 / 115000  # 1/kW
        c_tubes = 0.4
        return m_tubes * nominal_stack_power / 1000 + c_tubes  # W/m²/K

    def __calculate_surface_ratio_separator(self, nominal_stack_power) -> float:
        m_separator = -329 / 57500  # 1/kW
        c_separator = 8.52
        return (m_separator * nominal_stack_power / 1000 + c_separator)  * 10 ** (-6)  # m²/cm²

    def __calculate_max_water_flow_stack(self, electrolyzer_nominal_power):
        """ calculates max water flow the cooling pump can provide in dependency of nominal power of electrolyzer
        based on data from thesis: PEM-Elektrolyse-Systeme zur Anwendung in Power-to-Gas Anlagen"""
        # max water flow rate = 1.5 kg/s for a 1250 kW EL-Stack
        water_flow_rate = 11.9 / 1250  # 11.9 kg/s / 1250 kW
        el_nom_power_kW = electrolyzer_nominal_power / 1000  # W -> kW
        return water_flow_rate * el_nom_power_kW / Hydrogen.MOLAR_MASS_WATER  # mol/s

    def calculate_pump_power(self, water_flow_stack) -> None:
        relative_flow = water_flow_stack / self.__max_water_flow_stack
        pressure_loss = 1.985 * 10 ** (-4) * (relative_flow * 100) ** 2 * 10 ** (5) # N/m²
        volume_flow = water_flow_stack * Hydrogen.MOLAR_MASS_WATER / Hydrogen.DENSITY_WATER  # m³/s
        self.__pump.calculate_pump_power(volume_flow * pressure_loss)  # W
        self.__pump_power = self.__pump.get_pump_power()

    def get_temperature(self) -> float:
        return self.__temperature_stack_1

    def get_water_flow_stack(self) -> float:
        return self.__h2o_flow_stack

    def get_power_water_heating(self) -> float:
        return self.__heat_h2o

    def get_pump_power(self) -> float:
        return self.__pump_power

    def get_convection_heat(self) -> float:
        return self.__convection_heat



import math


class AlkalinePressureModel:    # TODO Check if assumption of equal pressures is correct/applicable

    def __init__(self):
        pass

    def get_sat_pressure_h2o(self, molarity: float, stack_temperature: float) -> float:
        # return self.__get_aqueous_vapour_pressure_koh(molarity, stack_temperature)
        pass

    def get_partial_pressure_h2(self, molarity: float, stack_temperature: float):
        return self.__get_aqueous_vapour_pressure_koh(molarity, stack_temperature)

    def get_partial_pressure_o2(self, molarity: float, stack_temperature: float):
        return self.__get_aqueous_vapour_pressure_koh(molarity, stack_temperature)

    def get_aqueous_vapour_pressure_koh(self, molarity: float, stack_temperature: float) -> float:
        return self.__get_aqueous_vapour_pressure_koh(molarity, stack_temperature)

    def __get_aqueous_vapour_pressure_koh(self, molarity: float, stack_temperature: float) -> float:
        return (stack_temperature ** -3.498) * math.exp(37.93 - (6426.32 / stack_temperature)) * \
               math.exp(0.016214 - 0.13802 * molarity + 0.19330 * molarity ** 0.5)

    def get_vapour_pressure_pure_water(self, stack_temperature: float) -> float:
        return self.__get_vapour_pressure_pure_water(stack_temperature)

    def __get_vapour_pressure_pure_water(self, stack_temperature: float) -> float:
        return (stack_temperature ** -3.4159) * math.exp(37.043 - (6275.7 / stack_temperature))



from simses.commons.state.technology.electrolyzer import ElectrolyzerState
from simses.technology.hydrogen.electrolyzer.degradation.cyclic.cyclic_degradation import \
    CyclicDegradationModel


class CyclicDegradationPemElMultiDimAnalytic(CyclicDegradationModel):

    def __init__(self):
        super().__init__()
        self.__resistance_increase = 0
        self.__exchange_current_dens_decrease = 0  # p.u.
        self.__current_zero_counter = 0  # counts timesteps since begin of current = 0

    def calculate_resistance_increase(self, time: float, state: ElectrolyzerState) -> None:
        """
        Calculation of resistance increase dependent on currentdensity provided to the electrolyzer cell
        recovery effect: all time exponential
        based on paper: "Polymer electrolyte membrane water electrolysis: Restraining degradation in the presence
        of fluctuating power"  by Rakousky, Christoph
        year: 2017
        :param time:
        :param state:
        :return: none
        """
        timestep = time - state.time
        current_dens = state.current_density
        if current_dens == 0:  # recovery effect: all time exponential
            self.__current_zero_counter += 1
            time = self.__current_zero_counter * timestep / 3600  # s -> h
            # self.__resistance_increase = - 1.968 * 10 ** (-4) * 0.4263 * np.exp(0.4263 * time) * timestep / 3600  # exponential recovery of total resistance
            self.__resistance_increase = (- 2 * 5.151 * 10 ** (-5) * time - 7.991 * 10 ** (-5)) * timestep / 3600  # quadratic recovery of total resistance
        elif 0 < current_dens < 1:
            self.__current_zero_counter = 0
            self.__resistance_increase = 0
        elif current_dens > 1:
            self.__current_zero_counter = 0
            self.__resistance_increase = 3.89 * 10 ** (-4) * (current_dens-1) * timestep / 3600  # p.u.
        else:
            self.__current_zero_counter = 0
            self.__resistance_increase = 0  # p.u.

    def calculate_exchange_current_dens_decrerase(self, state: ElectrolyzerState):
        """ No cyclic exchange_current_density_decrease """
        self.__exchange_current_dens_decrease = 0

    def get_resistance_increase(self) -> float:
        return self.__resistance_increase

    def get_exchange_current_dens_decrease(self) -> float:
        return self.__exchange_current_dens_decrease

    def reset(self, hydrogen_state: ElectrolyzerState) -> None:
        pass

    def close(self) -> None:
        pass

    # def calculate_resistance_increase(self, time: float, hydrogen_state: HydrogenState) -> None:
    #     """
    #     Calculation of resistance increase dependent on currentdensity provided to the electrolyzer cell
    #     recovery effect: first 6h exponential, afterwards linear
    #     based on paper: "Polymer electrolyte membrane water electrolysis: Restraining degradation in the presence
    #     of fluctuating power"  by Rakousky, Christoph
    #     year: 2017
    #     :param time:
    #     :param hydrogen_state:
    #     :return: none
    #     """
    #     time_passed = time - hydrogen_state.time
    #     current_dens = hydrogen_state.current_density_el
    #     if current_dens == 0:
    #         if time_passed / 3600 <= 6:  # recovery effect: exponential
    #             self.__current_zero_counter += 1
    #             time = self.__current_zero_counter * time_passed / 3600  # s -> h
    #             self.__resistance_increase = - 1.968 * 10 ** (-4) * 0.4263 * np.exp(0.4263 * time) * time_passed / 3600  # exponential recovery of total resistance
    #         else:  # recovery effect: linear, with slop of exponentail recovery at time=6h
    #             self.__resistance_increase = -1.0828763 * 10 ** (-3) * time_passed / 3600
    #     elif current_dens > 0 and current_dens <= 1:
    #         self.__current_zero_counter = 0
    #         self.__resistance_increase = 0  # p.u.
    #     elif current_dens >= 1 and current_dens < 2:
    #         self.__current_zero_counter = 0
    #         self.__resistance_increase = 2.083 * 10 ** (-4) * time_passed / 3600  # p.u.
    #     elif current_dens >= 2:
    #         self.__current_zero_counter = 0
    #         self.__resistance_increase = 3.89 * 10 ** (-4) * time_passed / 3600  # p.u.
    #     else:
    #         self.__current_zero_counter = 0
    #         self.__resistance_increase = 0  # p.u.






import math
from simses.commons.state.technology.lithium_ion import LithiumIonState
from simses.technology.lithium_ion.cell.type import CellType
from simses.technology.lithium_ion.degradation.calendar.calendar_degradation import \
    CalendarDegradationModel


class PanasonicNCACalendarDegradationModel(CalendarDegradationModel):

    __capacity_loss = 0
    __resistance_increase = 0

    __TEMP_REF = 298.15  # K
    __SOC_REF = 0.5  # pu
    __R = 8.3144598  # J/(K*mol)

    __K_REF_QLOSS = 0.0013  # pu*s^(-0.5)
    __A_CAL_QLOSS = 1.630861630861708  # fit parameter for voltage dependence - quadratic term
    __B_CAL_QLOSS = - 10.723723723724325  # fit parameter for voltage dependence - linear term
    __C_CAL_QLOSS = 18.351282051283220  # fit parameter for voltage dependence - constant term
    __EA_QLOSS = 32681  # J/mol

    __K_REF_RINC = 0.001211111111111  # pu*s^(-0.5)
    __A_RINC = 3.238585817688798 * 10 ** -8  # fit parameter for resistance increase - soc dependence - 4th order term
    __B_RINC = - 5.557556793759961 * 10 ** -6  # fit parameter for resistance increase - soc dependence - 3rd order term
    __C_RINC = 3.654685225712923 * 10 ** -4  # fit parameter for resistance increase - soc dependence - 2nd order term
    __D_RINC = - 0.008173737926571  # fit parameter for resistance increase - soc dependence - 1st order term
    __E_RINC = 0.466408440817863  # fit parameter for resistance increase - soc dependence - constant term
    __EA_RINC = 2.796216012248313 * 10 ** 4  # J/mol

    # Values based on MA Yuhui Geng (data originally from Peter Keil)

    def __init__(self, cell_type: CellType):
        super().__init__(cell_type)
        self.__capacity_loss_cal = cell_type.get_calendar_capacity_loss_start()
        self.__rinc_cal = 0

    def calculate_degradation(self, time: float, battery_state: LithiumIonState) -> None:
        time_passed = time - battery_state.time
        temp = battery_state.temperature
        # Open circuit voltage is used for calendar degradation model
        voltage = self._cell.get_open_circuit_voltage(battery_state) / self._cell.get_serial_scale()

        qloss: float = self.__capacity_loss_cal

        virtual_time = (qloss / (
                self.__K_REF_QLOSS * math.exp(-self.__EA_QLOSS / self.__R * (1 / temp - 1 / self.__TEMP_REF)) * (
                self.__A_CAL_QLOSS * voltage ** 2 + self.__B_CAL_QLOSS * voltage + self.__C_CAL_QLOSS))) ** 2  # days

        rel_capacity_loss = lambda time: self.__K_REF_QLOSS * math.exp(
            -self.__EA_QLOSS / self.__R * (1 / temp - 1 / self.__TEMP_REF)) * (
                                                 self.__A_CAL_QLOSS * voltage ** 2 + self.__B_CAL_QLOSS * voltage + self.__C_CAL_QLOSS) * time ** 0.5

        capacity_loss = rel_capacity_loss(virtual_time + time_passed / 86400) - self.__capacity_loss_cal

        self.__capacity_loss = capacity_loss * self._cell.get_nominal_capacity()  # Ah
        self.__capacity_loss_cal += capacity_loss

    def calculate_resistance_increase(self, time: float, battery_state: LithiumIonState) -> None:
        time_passed = time - battery_state.time
        soc = battery_state.soc
        temp = battery_state.temperature
        rinc_cal = self.__rinc_cal

        virtual_time = (rinc_cal / (
                self.__K_REF_RINC * math.exp(-self.__EA_RINC / self.__R * (1 / temp - 1 / self.__TEMP_REF)) * (
                self.__A_RINC * (soc*100) ** 4 + self.__B_RINC * (soc*100) ** 3 + self.__C_RINC *
                (soc*100) ** 2 + self.__D_RINC * soc * 100 + self.__E_RINC)))

        rel_resistance_increase = lambda time: self.__K_REF_RINC * math.exp(
            -self.__EA_RINC / self.__R * (1 / temp - 1 / self.__TEMP_REF)) * (
                                                       self.__A_RINC * (soc*100) ** 4 + self.__B_RINC * (soc*100) ** 3 +
                                                       self.__C_RINC * (soc*100) ** 2 + self.__D_RINC * (soc*100) + self.__E_RINC) \
                                               * time

        resistance_increase = rel_resistance_increase(virtual_time + time_passed / 86400) - self.__rinc_cal

        self.__rinc_cal += resistance_increase
        self.__resistance_increase = resistance_increase  # pu

    def get_degradation(self) -> float:
        return self.__capacity_loss

    def get_resistance_increase(self) -> float:
        return self.__resistance_increase

    def reset(self, battery_state: LithiumIonState) -> None:
        self.__capacity_loss_cal = 0
        self.__rinc_cal = 0

    def close(self) -> None:
        pass





class PowerDistributorState:
    """
    PowerDistributorStates contains information for power distribution algorithms.
    """

    def __init__(self, sys_id: int, soh: float = 0.0, soc: float = 0.0, storage_technology: str = None,
                 max_charge_power: float = 0.0, max_discharge_power: float = 0.0):
        """
        Constructor of PowerDistributorStates

        Parameters
        ----------

        sys_id :
            number of ac or dc system
        soh :
            state of health of a system
        soc :
            state of charge of a system
        storage_technology:
            type of storage technology
        """

        self.__soh: float = soh
        self.__id: int = sys_id
        self.__soc: float = soc
        self.__max_charge_power: float = max_charge_power
        self.__max_discharge_power: float = max_discharge_power
        self.__rebalance_charge_power: float = 0.0
        self.__rebalance_discharge_power: float = 0.0
        self.__time_delta: float = 0.0
        self.__capacity: float = 0.0
        self.__storage_technology: str = storage_technology

    @property
    def soh(self) -> float:
        return self.__soh

    @soh.setter
    def soh(self, value: float) -> None:
        self.__soh = value

    @property
    def sys_id(self) -> float:
        return self.__id

    @sys_id.setter
    def sys_id(self, value: float) -> None:
        self.__id = value

    @property
    def soc(self) -> float:
        return self.__soc

    @soc.setter
    def soc(self, value: float) -> None:
        self.__soc = value

    @property
    def time_delta(self) -> float:
        return self.__time_delta

    @time_delta.setter
    def time_delta(self, value: float) -> None:
        self.__time_delta = value

    @property
    def capacity(self) -> float:
        return self.__capacity

    @capacity.setter
    def capacity(self, value: float) -> None:
        self.__capacity = value

    @property
    def storage_technology(self) -> str:
        return self.__storage_technology

    @property
    def max_charge_power(self) -> float:
        return self.__max_charge_power

    @max_charge_power.setter
    def max_charge_power(self, value: float) -> None:
        self.__max_charge_power = value

    @property
    def max_discharge_power(self) -> float:
        return self.__max_discharge_power

    @max_discharge_power.setter
    def max_discharge_power(self, value: float) -> None:
        self.__max_discharge_power = value

    @property
    def rebalance_charge_power(self) -> float:
        return self.__rebalance_charge_power

    @rebalance_charge_power.setter
    def rebalance_charge_power(self, value: float) -> None:
        self.__rebalance_charge_power = value

    @property
    def rebalance_discharge_power(self) -> float:
        return self.__rebalance_discharge_power

    @rebalance_discharge_power.setter
    def rebalance_discharge_power(self, value: float) -> None:
        self.__rebalance_discharge_power = value

    @staticmethod
    def sort_by_soc(data: list, descending: bool = False) -> None:
        """
        In-place sorting of list with PowerDistributorState objects by soc (ascending by default)

        Parameters
        ----------
        descending :
            reverse sorting of data list, default: False
        data :
            list of TimeValue objects

        Returns
        -------

        """
        data.sort(key=lambda x: x.soc, reverse=descending)

    @staticmethod
    def sort_by_soh(data: list, descending: bool = True) -> None:
        """
        In-place sorting of list with PowerDistributorState objects by soh (descending by default)

        Parameters
        ----------
        descending :
            reverse sorting of data list, default: True
        data :
            list of TimeValue objects

        Returns
        -------

        """
        data.sort(key=lambda x: x.soh, reverse=descending)



import multiprocessing
import os
import time
from abc import abstractmethod, ABC
from configparser import ConfigParser
from datetime import datetime
from multiprocessing import Queue

from simses.commons.console_printer import ConsolePrinter
from simses.commons.utils.utilities import format_float
from .batch_simulation import BatchSimulation


class BatchProcessing(ABC):

    """
    BatchProcessing delivers the necessary handling for running multiple parallel SimSES simulations and distributes
    the configuration to each process. It supervises all processes and starts as many processes as there are cores
    available. If you run more simulations than cores available, it waits until a simulation has finished and fills
    the gap with new simulation processes.
    """

    UPDATE: int = 1  # s

    def __init__(self, path: str = os.getcwd(),  do_simulation: bool = True, do_analysis: bool = True):
        self.__do_simulation: bool = do_simulation
        self.__do_analysis: bool = do_analysis
        self.__max_parallel_processes: int = max(1, multiprocessing.cpu_count() - 1)
        self.__path: str = os.path.join(path, 'results').replace('\\','/') + '/'
        print('Results will be stored in ' + self.__path)
        self.__result_paths: [str] = list()

    @abstractmethod
    def _setup_config(self) -> dict:
        """
        Setting up the necessary configuration for multiple simulations

        Returns
        -------
        dict:
            a dictionary of configs
        """
        pass

    def _analysis_config(self) -> ConfigParser:
        """
        Setting up the analysis configuration

        Returns
        -------
        ConfigParser:
            config for analysis
        """
        return ConfigParser()

    @abstractmethod
    def clean_up(self) -> None:
        pass

    def get_result_paths(self) -> [str]:
        out: [str] = list()
        for path in self.__result_paths:
            out.append(self.__get_latest_path_from(path))
        return out

    def __get_latest_path_from(self, path: str) -> str:
        result_dirs = list()
        tmp_dirs = os.listdir(path)
        for directory in tmp_dirs:
            if os.path.isdir(path + '/' + directory):
                result_dirs.append(path + '/' + directory + '/')
        return max(result_dirs, key=os.path.getmtime)

    def run(self):
        config_set: dict = self._setup_config()
        analysis_config: ConfigParser = self._analysis_config()
        print(str(len(config_set)) + ' simulations configured')
        printer_queue: Queue = Queue(maxsize=len(config_set) * 2)
        printer: ConsolePrinter = ConsolePrinter(printer_queue)
        jobs: [BatchSimulation] = list()
        for key, value in config_set.items():
            self.__result_paths.append(self.__path + key)
            jobs.append(BatchSimulation(config_set={key: value}, printer_queue=printer_queue, path=self.__path,
                                        do_simulation=self.__do_simulation, do_analysis=self.__do_analysis,
                                        analysis_config=analysis_config))
        printer.start()
        started: [BatchSimulation] = list()
        start = time.time()
        job_count: int = len(jobs)
        count: int = 0
        for job in jobs:
            job.start()
            started.append(job)
            count += 1
            print('\rStarting simulation ' + str(count) + '/' + str(job_count) + ' @ ' + str(datetime.now()))
            self.__check_running_process(started)
        for job in jobs:
            job.join()
        duration: float = (time.time() - start) / 60.0
        print('\r\n' + type(self).__name__ + ' finished ' + str(job_count) + ' simulations in ' + format_float(duration) + ' min '
              '(' + format_float(duration / job_count) + ' min per simulation)')
        printer.stop_immediately()

    def __check_running_process(self, processes: [multiprocessing.Process]) -> None:
        while True:
            running_jobs: int = 0
            for process in processes:
                if process.is_alive():
                    running_jobs += 1
            if running_jobs < self.__max_parallel_processes:
                break
            time.sleep(self.UPDATE)



from simses.commons.config.simulation.general import GeneralSimulationConfig
from simses.system.dc_coupling.dc_coupler import DcCoupling
from simses.system.dc_coupling.generation.no_dc_generation import NoDcGeneration
from simses.system.dc_coupling.load.dc_bus_charging_profile import DcBusChargingProfile


class BusChargingProfileDcCoupling(DcCoupling):

    def __init__(self, config: GeneralSimulationConfig, capacity: float, file_name: str):
        super().__init__(DcBusChargingProfile(config, capacity, file_name), NoDcGeneration())



import numpy
import time as calctime
from simses.analysis.data.energy_management import EnergyManagementData
from simses.analysis.data.system import SystemData
from simses.analysis.evaluation.plotting.axis import Axis
from simses.analysis.evaluation.plotting.plotly_plotting import PlotlyPlotting
from simses.analysis.evaluation.plotting.plotter import Plotting
from simses.analysis.evaluation.result import EvaluationResult, Description, Unit
from simses.analysis.evaluation.technical.technical_evaluation import TechnicalEvaluation
from simses.analysis.utils import get_max_for
from simses.commons.config.analysis.general import GeneralAnalysisConfig
from simses.commons.config.simulation.energy_management import EnergyManagementConfig
from simses.commons.state.energy_management import EnergyManagementState
from simses.commons.state.system import SystemState


class SiteLevelEvaluation(TechnicalEvaluation):

    title_power = 'Site Level Power'
    title_idm = 'Intraday Power'

    def __init__(self, data: SystemData, energy_management_data: EnergyManagementData, config: GeneralAnalysisConfig,
                 energy_management_config: EnergyManagementConfig, path: str):
        super().__init__(energy_management_data, config)
        self.__system_data: SystemData = data
        self.__max_power: float = energy_management_config.max_power
        title_extension: str = ' for system ' + self.get_data().id
        self.title_power += title_extension
        self.title_idm += title_extension
        self.__result_path = path
        self.__max_power_timeseries = energy_management_data.peakshaving_limit

    def evaluate(self):
        power_above_peak = self.power_above_peak
        energy_events_above_peak = self.energy_events_above_peak(power_above_peak)
        self.append_result(EvaluationResult(Description.Technical.MAX_GRID_POWER, Unit.WATT, self.max_grid_power))
        self.append_result(EvaluationResult(Description.Technical.POWER_ABOVE_PEAK_MAX, Unit.WATT, self.max_power_above_peak(power_above_peak)))
        self.append_result(EvaluationResult(Description.Technical.POWER_ABOVE_PEAK_AVG, Unit.WATT, self.get_average_power_above_peak_from(power_above_peak)))
        self.append_result(EvaluationResult(Description.Technical.ENERGY_ABOVE_PEAK_MAX, Unit.KWH, self.get_max_energy_event_above_peak_from(energy_events_above_peak)))
        self.append_result(EvaluationResult(Description.Technical.ENERGY_ABOVE_PEAK_AVG, Unit.KWH, self.get_average_energy_event_above_peak_from(energy_events_above_peak)))
        self.append_result(EvaluationResult(Description.Technical.SELF_CONSUMPTION_RATE, Unit.PERCENTAGE, self.self_consumption_rate))
        self.append_result(EvaluationResult(Description.Technical.SELF_SUFFICIENCY, Unit.PERCENTAGE, self.self_sufficiency))
        self.append_result(EvaluationResult(Description.Technical.NUMBER_ENERGY_EVENTS, Unit.NONE, len(energy_events_above_peak)))
        self.append_result(EvaluationResult(Description.Technical.ENERGY_IDM_BOUGHT, Unit.KWH, self.intraday_energy_bought))
        self.append_result(EvaluationResult(Description.Technical.ENERGY_IDM_SOLD, Unit.KWH, self.intraday_energy_sold))
        self.print_results()

    def plot(self) -> None:
        self.site_level_power_plotting()
        self.intraday_power_plotting()

    def site_level_power_plotting(self) -> None:
        ems_data: EnergyManagementData = self.get_data()
        system_data: SystemData = self.__system_data
        plot: Plotting = PlotlyPlotting(title=self.title_power, path=self.__result_path)
        time = Plotting.format_time(system_data.time)
        xaxis: Axis = Axis(data=time, label=SystemState.TIME)
        yaxis: [Axis] = list()
        yaxis.append(Axis(system_data.power, label=SystemState.AC_POWER_DELIVERED, color=PlotlyPlotting.Color.AC_POWER_BLUE,
                          linestyle=PlotlyPlotting.Linestyle.SOLID))
        yaxis.append(Axis(ems_data.load_power, label=EnergyManagementState.LOAD_POWER,
                          color=PlotlyPlotting.Color.DC_POWER_GREEN,
                          linestyle=PlotlyPlotting.Linestyle.DASHED))
        # yaxis.append(Axis(self.__energy_management_data.pv_power, label=EnergyManagementState.PV_POWER, color=Plotting.Color.BLUE,
        #                  linestyle=Plotting.Linestyle.DASH_DOT))
        # grid_power = [load + battery - pv for load, battery, pv in zip(data.power, self.__energy_management_data.load_power, self.__energy_management_data.pv_power)]
        grid_power = system_data.power + ems_data.load_power - ems_data.pv_power
        yaxis.append(Axis(grid_power, label='Total Site Power in W', color=PlotlyPlotting.Color.POWER_YELLOW,
                          linestyle=PlotlyPlotting.Linestyle.DOTTED))
        yaxis.append(Axis(system_data.soc, label=SystemState.SOC, color=PlotlyPlotting.Color.SOC_BLUE,
                          linestyle=PlotlyPlotting.Linestyle.SOLID))
        yaxis.append(
            Axis(self.__max_power_timeseries, label='Peak power allowed in W', color=PlotlyPlotting.Color.MAGENTA,
                 linestyle=PlotlyPlotting.Linestyle.DASHED))
        plot.lines(xaxis, yaxis, [3])
        self.extend_figures(plot.get_figures())

    def intraday_power_plotting(self) -> None:
        try:
            ems_data: EnergyManagementData = self.get_data()
            if sum(abs(ems_data.idm_power)) > 0.0:
                plot: Plotting = PlotlyPlotting(title=self.title_idm, path=self.__result_path)
                time = Plotting.format_time(ems_data.time)
                xaxis: Axis = Axis(data=time, label=SystemState.TIME)
                yaxis: [Axis] = [Axis(ems_data.idm_power, label=EnergyManagementState.IDM_POWER)]
                plot.lines(xaxis, yaxis)
                self.extend_figures(plot.get_figures())
        except KeyError as err:
            self.__log.warn(err)

    @property
    def intraday_energy_bought(self) -> float:
        ems_data: EnergyManagementData = self.get_data()
        charge_energy = ems_data.idm_power[:].copy() * ems_data.convert_watt_to_kWh
        return abs(sum(charge_energy[charge_energy > 0.0]))

    @property
    def intraday_energy_sold(self) -> float:
        ems_data: EnergyManagementData = self.get_data()
        discharge_energy = ems_data.idm_power[:].copy() * ems_data.convert_watt_to_kWh
        return abs(sum(discharge_energy[discharge_energy < 0.0]))

    @property
    def grid_power(self) -> numpy.ndarray:
        data: EnergyManagementData = self.get_data()
        grid_power = self.__system_data.power + data.load_power - data.pv_power
        return grid_power

    @property
    def max_grid_power(self) -> float:
        return max(0.0, get_max_for(self.grid_power))

    @property
    def power_above_peak(self) -> numpy.ndarray:
        grid_power = self.grid_power
        power_above_peak = grid_power - self.__max_power_timeseries
        power_above_peak[power_above_peak < 0.0] = 0.0
        return power_above_peak

    @staticmethod
    def max_power_above_peak(power_above_peak) -> float:
        return max(0.0, get_max_for(power_above_peak))

    @staticmethod
    def get_average_power_above_peak_from(power_above_peak) -> float:
        power = power_above_peak[power_above_peak > 0.0]
        if len(power) == 0:
            return 0.0
        return sum(power) / len(power)

    def energy_events_above_peak(self, power_above_peak) -> [float]:
        energy_above_peak = power_above_peak * self.get_data().convert_watt_to_kWh
        energy_events: [float] = list()
        energy_event: float = 0.0
        for energy in energy_above_peak:
            if energy_event > 1e-5 and energy == 0.0:
                energy_events.append(energy_event)
                energy_event = 0.0
            else:
                energy_event += energy
        return energy_events

    @staticmethod
    def get_max_energy_event_above_peak_from(energy_events: [float]) -> float:
        if not energy_events:
            return 0.0
        return max(energy_events)

    @staticmethod
    def get_average_energy_event_above_peak_from(energy_events: [float]) -> float:
        if not energy_events:
            return 0.0
        return sum(energy_events) / len(energy_events)

    @property
    def self_consumption_rate(self) -> float:
        ems_data: EnergyManagementData = self.get_data()
        system_data: SystemData = self.__system_data
        grid_exchange: numpy.ndarray = ems_data.pv_power - system_data.power - ems_data.load_power
        grid_feed_in: float = sum(grid_exchange[grid_exchange > 0.0])
        pv_generation: float = sum(ems_data.pv_power)
        if pv_generation > 0.0:
            self_consumption: float = 100.0 * (1.0 - grid_feed_in / pv_generation)
        else:
            self_consumption: str = '-No PV- 0'
        return self_consumption

    @property
    def self_sufficiency(self) -> float:
        ems_data: EnergyManagementData = self.get_data()
        system_data: SystemData = self.__system_data
        grid_exchange: numpy.ndarray = ems_data.pv_power - system_data.power - ems_data.load_power
        grid_purchase: float = sum(-grid_exchange[grid_exchange < 0.0])
        load: float = sum(ems_data.load_power)
        pv_generation: float = sum(ems_data.pv_power)
        if pv_generation > 0.0:
            self_sufficiency_value: float = 100.0 * (1.0 - grid_purchase / load)
        else:
            self_sufficiency_value: str = '-No PV- 0'
        return self_sufficiency_value



from abc import ABC, abstractmethod
from simses.commons.utils.utilities import format_float


class State(ABC):

    """
    Basic abstract class for all states in SimSES. States in SimSES are date objects 1) transfering data from one
    object to another and 2) documenting results of simulation. Those written state files are used for the analysis
    of the simulation. There are no cached results in SimSES, all states are written immediately after each simulation
    timestep. The underlying state object in this class is a plain dictonary.
    """

    TIME = 'Time in s'

    def __init__(self):
        self.__state = dict()

    def _initialize(self) -> None:
        """
        Initializes all keys of state class with 0.

        Requirement: Keys have to be written in CAPITAL_LETTERS.

        Returns
        -------
        None:
            None
        """
        for attr in self.__class_attributes():
            self.init_variable(getattr(self, attr))

    @classmethod
    def __class_attributes(cls) -> [str]:
        return [attr for attr in dir(cls) if attr.isupper() and not attr.startswith("_")
                and not callable(getattr(cls, attr))]

    @classmethod
    def header(cls) -> [str]:
        """
        Returns a list of all state keys

        Returns
        -------

        """
        keys: [str] = list()
        for attr in cls.__class_attributes():
            keys.append(getattr(cls, attr))
        return keys
        # return list(self.__state.keys())

    def to_export(self) -> dict:
        """
        Returns a dictionary with a list of all values

        Returns
        -------
        dict:
            values combined with state.
        """
        export_dict: dict = dict()
        key: str = type(self).__name__
        export_dict[key] = {}
        export_dict[key]['DATA'] = self.to_list
        export_dict[key]['TIME'] = self.time
        export_dict[key]['ID'] = self.id
        return export_dict

    @property
    @abstractmethod
    def id(self) -> str:
        pass

    def __str__(self) -> str:
        res = ''
        for val in list(self.__state.values()):
            res += format_float(val) + ','
        return res

    def __repr__(self):
        return self.__str__()

    @property
    def to_list(self) -> list:
        """
        Returns a list of all values of self.state

        Returns
        -------
        list
        """
        return list(self.__state.values())

    @property
    def time(self) -> float:
        """
        Time in s

        Returns
        -------

        """
        return self.get(self.TIME)

    @time.setter
    def time(self, value: float) -> None:
        self.set(self.TIME, value)

    def get(self, key: str) -> float:
        """
        Returns value for key

        Parameters
        ----------
        key : str

        Returns
        -------
        value as float
        """
        return self.__state[key]

    def set(self, key: str, value: float) -> None:
        """
        Sets value to variable of key

        Parameters
        ----------
        key : str
        value : float

        Returns
        -------
        None
        """
        self.__state[key] = value

    def init_variable(self, key: str) -> None:
        """
        Intitalize variable of key with 0.

        Parameters
        ----------
        key : str

        Returns
        -------
        None
        """
        self.set(key, 0)

    def __check_instance_of(self, state) -> None:
        if not isinstance(state, self.__class__):
            raise Exception('State ' + state.__class__.__name__ + ' is not an instance of ' + self.__class__.__name__ +
                            '. Please check state transfer.')

    def add(self, state) -> None:
        """
        Adding all items of state to self.state

        Parameters
        ----------
        state : State

        Returns
        -------
            None
        """
        self.__check_instance_of(state)
        for key in self.__state.keys():
            try:
               self.__state[key] += state.get(key)
            except TypeError as err:
                raise TypeError('Key: ' + key, err)

    def set_all(self, state) -> None:
        """
        Setting all items of state to self.state

        Parameters
        ----------
        state : State

        Returns
        -------
            None
        """
        self.__check_instance_of(state)
        for key in self.__state.keys():
            self.__state[key] = state.get(key)

    def divide_by(self, divisor: float, key=None) -> None:
        """
        Divide all values of self.state by divisor. If a key is provided, only the value for the key is divided.

        Parameters
        ----------
        divisor : float
        key : str

        Returns
        -------
            None
        """
        if divisor == 0:
            raise Exception('Divison by zero')
        if key is None:
            for key in self.__state.keys():
                self.__state[key] /= divisor
        else:
            self.__state[key] /= divisor

    @classmethod
    def sum_reverse(cls, key: str, states: []) -> float:
        """
        Calculation the inverted sum of inverse values (like parallel resistances).

        Parameters
        ----------
        key : str
            Key of the state.
        states : list
            List of states.

        Returns
        -------
        float:
            Summed up value.

        """
        value = 0
        for state in states:
            value += 1 / state.get(key)
        return 1 / value

    @classmethod
    @abstractmethod
    def sum_parallel(cls, states: []):
        """
        Classmethod to calculate the combined state over several states connected in parallel.

        Parameters
        ----------
        states: List
            List of State which are connected in parallel.

        Returns
        -------
        state:
            Combined State over the parallel States.
        """
        pass

    @classmethod
    @abstractmethod
    def sum_serial(cls, states: []):
        """
        Classmethod to calculate the combined state over several states connected in serial.

        Parameters
        ----------
        states: List
            List of State which are connected in serial.

        Returns
        -------
        state:
            Combined State over the serial States.
        """
        pass



from abc import ABC, abstractmethod


class AmbientThermalModel(ABC):

    """
    AmbientThermalModel provides a temperature of the ambient for the system thermal model calculations.
    """

    def __init__(self):
        super().__init__()
        pass

    @abstractmethod
    def get_temperature(self, time) -> float:
        """
        Returns the ambient temperature

        Parameters
        -------
        time : current simulation time

        Returns
        -------
        float
            ambient temperature in Kelvin
        """
        pass

    @abstractmethod
    def get_initial_temperature(self) -> float:
        """
        Returns the ambient temperature

        Parameters
        -------
        time : current simulation time

        Returns
        -------
        float
            ambient temperature in Kelvin
        """
        pass

    @abstractmethod
    def create_instance(self):
        """
        reinstantiates the AmbientThermalModel
        :return: AmbientThermalModel
        """
        pass

    @abstractmethod
    def close(self):
        """
        closes all open resources in ambient thermal model

        Parameters
        ----------

        Returns
        -------

        """
        pass



from simses.commons.config.simulation.redox_flow import RedoxFlowConfig
from simses.commons.log import Logger
from simses.commons.state.technology.redox_flow import RedoxFlowState
from simses.technology.redox_flow.pump_algorithm.abstract_pump_algorithm import PumpAlgorithm
from simses.technology.redox_flow.stack.electrolyte.abstract_electrolyte import ElectrolyteSystem
from simses.technology.redox_flow.stack.abstract_stack import StackModule


class RedoxControlSystem:
    """RedoxControlSystem class for redox flow batteries. It contains queries to check whether the operation
    conditions are met."""

    def __init__(self, stack_module: StackModule, electrolyte_system: ElectrolyteSystem, pump_algorithm: PumpAlgorithm,
                 redox_flow_config: RedoxFlowConfig):
        self.__log: Logger = Logger(type(self).__name__)
        self.__stack_module: StackModule = stack_module
        self.__electrolyte_system: ElectrolyteSystem = electrolyte_system
        self.__pump_algorithm: PumpAlgorithm = pump_algorithm
        self.__soc_min = redox_flow_config.min_soc
        self.__soc_max = redox_flow_config.max_soc
        self.__FARADAY = self.__electrolyte_system.FARADAY

    def check_temperature(self, redox_flow_state: RedoxFlowState) -> bool:
        """
        Checks if the temperature of the electrolyte is within the operation range.

        Parameters
        ----------
        redox_flow_state : RedoxFlowState
            Current state of redox_flow.

        Returns
        -------
        bool:
            True if the temperature is within the operation range.
        """
        return (self.__electrolyte_system.get_min_temperature() <= redox_flow_state.temperature <=
                self.__electrolyte_system.get_max_temperature())

    def check_voltage_in_range(self, redox_flow_state: RedoxFlowState) -> bool:
        """
        Checks if voltage is in range between maximal and minimal stack module voltage.

        Parameters
        ----------
        redox_flow_state : RedoxFlowState
            Current state of redox_flow.

        Returns
        -------
        bool:
            True if voltage is between max. and min. stack module voltage.
        """
        return (self.__stack_module.get_min_voltage() <= redox_flow_state.voltage <=
                self.__stack_module.get_max_voltage())

    def voltage_in_range(self, redox_flow_state: RedoxFlowState) -> float:
        """
        If the voltage is outside of the maximal or minimal stack module voltage, then it is set to the maximal or
        minimal value.

        Parameters
        ----------
        redox_flow_state : RedoxFlowState
            Current state of redox_flow.

        Returns
        -------
        float:
            Voltage in V.
        """
        voltage_target = redox_flow_state.voltage
        if ((voltage_target > redox_flow_state.open_circuit_voltage) and (voltage_target >
            self.__stack_module.get_max_voltage())):
            voltage = max(redox_flow_state.open_circuit_voltage, self.__stack_module.get_max_voltage())
        elif ((voltage_target < redox_flow_state.open_circuit_voltage) and (voltage_target <
              self.__stack_module.get_min_voltage())):
            voltage = min(redox_flow_state.open_circuit_voltage, self.__stack_module.get_min_voltage())
        else:
            voltage = max(min(voltage_target, self.__stack_module.get_max_voltage()),
                          self.__stack_module.get_min_voltage())
        return voltage

    def check_soc_in_range(self, redox_flow_state: RedoxFlowState, delta_soc: float, delta_soc_stack: float) -> bool:
        """
        Checks if the state-of-charge (SOC) is to high or to low for charging or discharging.

        Parameters
        ----------
        redox_flow_state : RedoxFlowState
            Current state of redox_flow.
        delta_soc: float
            State-of-charge in p.u.
        delta_soc_stack: float
            State-of-charge in the stack in p.u.


        Returns
        -------
        bool:
            True if SOC of the system and in the stack is in range for charging or discharging.
        """
        if redox_flow_state.is_charge and (((redox_flow_state.soc + delta_soc) >= self.__soc_max) or
                                           ((redox_flow_state.soc_stack + delta_soc_stack) >= 1)):
            return False
        elif not redox_flow_state.is_charge and (((redox_flow_state.soc + delta_soc) <= self.__soc_min) or
                                                 ((redox_flow_state.soc_stack + delta_soc_stack) <= 0)):
            return False
        else:
            return True

    def check_current_in_range(self, redox_flow_state: RedoxFlowState, time: float) -> bool:
        """
        Checks if the current is in range. The maximal and minimal current are defined to have enough reactants at the
        current flow rate or in the stack electrolyte volume.

        Parameters
        ----------
        redox_flow_state : RedoxFlowState
            Current state of redox_flow.
        time : float
            Current simulation time in s.

        Returns
        -------
        bool:
            True if current is in range.
        """
        if redox_flow_state.current > self.get_max_current(redox_flow_state, time):
            self.__log.warn('Flow Rate for charging is to low')
            return False
        elif redox_flow_state.current < self.get_min_current(redox_flow_state, time):
            self.__log.warn('Flow Rate for discharging is to low')
            return False
        else:
            return True

    def get_min_current(self, redox_flow_state: RedoxFlowState, time: float) -> float:
        """
        Determines the minimal faraday current of a stack module (maximal discharge current) based on the flow rate.
        If the flow rate is 0, the electrolyte volume in the stack modules and the still usable concentration of
        reactants is used for calculation.

        Parameters
        ----------
        redox_flow_state : RedoxFlowState
            Current state of redox_flow.
        time : float
            Current simulation time in s.

        Returns
        -------
        float:
            Minimal faraday current (=max discharge current) in A.
        """
        soc_stack = redox_flow_state.soc_stack
        flow_rate = min(self.__pump_algorithm.get_flow_rate_anolyte(redox_flow_state),
                        self.__pump_algorithm.get_flow_rate_catholyte(redox_flow_state))
        time_step = time - redox_flow_state.time
        if flow_rate == 0.0:
            min_current = (-soc_stack * self.__FARADAY * self.__electrolyte_system.get_vanadium_concentration() *
                           self.__stack_module.get_stacks_electrolyte_volume() / time_step /
                           (self.__stack_module.get_serial_scale() * self.__stack_module.get_cell_per_stack()))
        else:
            min_current = (-soc_stack * self.__FARADAY * self.__electrolyte_system.get_vanadium_concentration() *
                           flow_rate / (self.__stack_module.get_serial_scale() *
                           self.__stack_module.get_cell_per_stack()))
        return min_current

    def get_max_current(self, redox_flow_state: RedoxFlowState, time: float) -> float:
        """
        Determines the maximal faraday current of a stack module (maximal charge current) based on the flow rate.
        If the flow rate is 0, the electrolyte volume in the stack modules and the still usable concentration of
        reactants is used for calculation.

        Parameters
        ----------
        redox_flow_state : RedoxFlowState
            Current state of redox_flow.
        time : float
            Current simulation time in s.

        Returns
        -------
        float:
            Maximal faraday current (=max charge current) in A.
        """
        soc_stack = redox_flow_state.soc_stack
        flow_rate = min(self.__pump_algorithm.get_flow_rate_anolyte(redox_flow_state),
                        self.__pump_algorithm.get_flow_rate_catholyte(redox_flow_state))
        time_step = time - redox_flow_state.time
        if flow_rate == 0.0:
            max_current = ((1 - soc_stack) * self.__FARADAY * self.__electrolyte_system.get_vanadium_concentration() *
                           self.__stack_module.get_stacks_electrolyte_volume() / time_step /
                           (self.__stack_module.get_serial_scale() * self.__stack_module.get_cell_per_stack()))
        else:
            max_current = ((1 - soc_stack) * self.__FARADAY * self.__electrolyte_system.get_vanadium_concentration() *
                           flow_rate / (self.__stack_module.get_serial_scale() *
                           self.__stack_module.get_cell_per_stack()))
        return max_current

    def battery_fulfillment_calc(self, redox_flow_state: RedoxFlowState) -> None:
        """
        Calculates the battery fulfillment [0, 1].

        Parameters
        ----------
        redox_flow_state : RedoxFlowState
            Current state of redox_flow.

        Returns
        -------
        """
        power_is = redox_flow_state.power
        power_target = redox_flow_state.power_in
        if abs(power_is - power_target) < 1e-8:
            redox_flow_state.fulfillment = 1.0
        elif power_target == 0:
            self.__log.error('Power should be 0, but is ' + str(power_is) + ' A. Check BMS function.')
            redox_flow_state.fulfillment = 0.0
        else:
            redox_flow_state.fulfillment = abs(power_is / power_target)
            if redox_flow_state.fulfillment < 0 or redox_flow_state.fulfillment > 1:
                self.__log.error('Fulfillment should be between 0 and 1, but is ' +
                                 str(redox_flow_state.fulfillment) + '. Check BMS functions.')

    def set_max_power(self, redox_flow_state: RedoxFlowState) -> None:
        max_power: float = self.__stack_module.get_max_power()
        redox_flow_state.max_charge_power = 0.0 if redox_flow_state.soc > self.__soc_max else max_power
        redox_flow_state.max_discharge_power = 0.0 if redox_flow_state.soc < self.__soc_min else max_power

    def close(self):
        """Closing all resources in RedoxControlSystem."""
        self.__log.close()
        self.__stack_module.close()
        self.__electrolyte_system.close()
        self.__pump_algorithm.close()



import math

import pandas as pd
import scipy.interpolate
from numpy import asarray

from simses.commons.config.data.battery import BatteryDataConfig
from simses.commons.config.simulation.battery import BatteryConfig
from simses.commons.log import Logger
from simses.commons.state.technology.lithium_ion import LithiumIonState
from simses.technology.lithium_ion.cell.electric.properties import ElectricalCellProperties
from simses.technology.lithium_ion.cell.format.abstract import CellFormat
from simses.technology.lithium_ion.cell.format.round_18650 import RoundCell18650
from simses.technology.lithium_ion.cell.thermal.properties import ThermalCellProperties
from simses.technology.lithium_ion.cell.type import CellType


class SanyoNMC(CellType):
    """An NMC (Sanyo UR18650E) is a special cell type and inherited by CellType"""

    """Source: Schmalstieg, J., Käbitz, S., Ecker, M., & Sauer, D. U. (2014). 
    A holistic aging model for Li (NiMnCo) O2 based 18650 lithium-ion batteries. 
    Journal of Power Sources, 257, 325-334."""

    __SOC_HEADER = 'SOC'
    __SOC_IDX = 0
    __OCV_IDX = 1
    __TEMP_IDX = 1
    __C_RATE_IDX = 0
    __ETA_IDX = 1

    __CELL_VOLTAGE = 3.6  # V
    __CELL_CAPACITY = 2.05  # Ah
    __MAX_VOLTAGE: float = 4.2  # V
    __MIN_VOLTAGE: float = 2.5  # V
    __MIN_TEMPERATURE: float = 273.15  # K
    __MAX_TEMPERATURE: float = 313.15  # K
    __MAX_CHARGE_RATE: float = 1.0  # 1/h
    __MAX_DISCHARGE_RATE: float = 3.0  # 1/h
    __SELF_DISCHARGE_RATE: float = 0.0  # X.X%-soc per day, e.g., 0.015 for 1.5% SOC loss per day
    __MASS: float = 0.046  # kg per cell
    __SPECIFIC_HEAT: float = 965  # J/kgK
    __CONVECTION_COEFFICIENT: float = 15  # W/m2K

    __COULOMB_EFFICIENCY: float = 1.0  # p.u.

    __ELECTRICAL_PROPS: ElectricalCellProperties = ElectricalCellProperties(__CELL_VOLTAGE, __CELL_CAPACITY,
                                                                            __MIN_VOLTAGE, __MAX_VOLTAGE,
                                                                            __MAX_CHARGE_RATE, __MAX_DISCHARGE_RATE,
                                                                            __SELF_DISCHARGE_RATE, __COULOMB_EFFICIENCY)
    __THERMAL_PROPS: ThermalCellProperties = ThermalCellProperties(__MIN_TEMPERATURE, __MAX_TEMPERATURE, __MASS,
                                                                   __SPECIFIC_HEAT, __CONVECTION_COEFFICIENT)
    __CELL_FORMAT: CellFormat = RoundCell18650()

    def __init__(self, voltage: float, capacity: float, soh: float, battery_config: BatteryConfig,
                 battery_data_config: BatteryDataConfig):
        super().__init__(voltage, capacity, soh, self.__ELECTRICAL_PROPS, self.__THERMAL_PROPS, self.__CELL_FORMAT,
                         battery_config)
        self.__log: Logger = Logger(type(self).__name__)
        rint_file: str = battery_data_config.nmc_sanyo_rint_file
        internal_resistance = pd.read_csv(rint_file)  # Ohm
        self.__log.warn('No Rint data for this cell. internal Resistance from Molicel is assumed.')
        soc_arr = internal_resistance.iloc[:, self.__SOC_IDX]
        temp_arr = internal_resistance.iloc[:4, self.__TEMP_IDX]
        rint_mat_ch = internal_resistance.iloc[:, 2]
        rint_mat_dch = internal_resistance.iloc[:, 5]
        self.__rint_interp1d_ch = scipy.interpolate.interp1d(soc_arr, rint_mat_ch, kind='linear')
        self.__rint_interp1d_dch = scipy.interpolate.interp1d(soc_arr, rint_mat_dch, kind='linear')
        self.__log.debug('soc arr size: ' + str(len(soc_arr)) + ', temp arr size: ' + str(
            len(temp_arr.T)) + ', rint ch mat size: ' + str(asarray(rint_mat_ch).shape))
        self.__log.debug('soc arr size: ' + str(len(soc_arr)) + ', temp arr size: ' + str(
            len(temp_arr.T)) + ', rint dch mat size: ' + str(asarray(rint_mat_dch).shape))

    def get_open_circuit_voltage(self, battery_state: LithiumIonState) -> float:
        '''Parameters build with ocv fitting'''
        a1 = -7.7487
        a2 = -0.0974
        a3 = 1.2023
        a4 = 3.9977
        b1 = -0.1714
        b2 = 2.6526
        k0 = 2.3885
        k1 = 2.1430
        k2 = -0.6287
        k3 = -1.6708
        k4 = 1.6161
        k5 = 0.7234
        soc = battery_state.soc

        ocv = k0 + \
              k1 / (1 + math.exp(a1 * (soc - b1))) + \
              k2 / (1 + math.exp(a2 * (soc - b2))) + \
              k3 / (1 + math.exp(a3 * (soc - 1))) +\
              k4 / (1 + math.exp(a4 * soc)) +\
              k5 * soc

        return ocv * self.get_serial_scale()

    def get_internal_resistance(self, battery_state: LithiumIonState) -> float:
        # internal resistance from Molicel!
        soc = battery_state.soc
        soc = self.check_soc_range(soc)
        if battery_state.is_charge:
            # internal resistance for charge
            res = self.__rint_interp1d_ch(soc)
            self.__log.debug('res charging: ' + str(res / self.get_parallel_scale() * self.get_serial_scale()))
        else:
            # internal resistance for discharge
            res = self.__rint_interp1d_dch(soc)
            self.__log.debug('res discharging: ' + str(res / self.get_parallel_scale() * self.get_serial_scale()))
        return float(res) / self.get_parallel_scale() * self.get_serial_scale()

    def close(self):
        self.__log.close()



from abc import ABC, abstractmethod

from simses.commons.state.technology.storage import StorageTechnologyState


class CycleDetector(ABC):

    def __init__(self):
        pass

    @abstractmethod
    def cycle_detected(self, time: float, state: StorageTechnologyState) -> bool:
        """
        Cycle Detector. Returns true if the sign (charging to discharge and vice versa) changes or the SOC reaches the
        maximum or minimum or the end of the simulation is reached

        Parameters
        ----------
        time : current time of the simulation
        state : LithiumIonState, current BatteryState of the lithium_ion state

        Returns
        -------
        bool:
            returns true if a cycle is detected

        """
        pass

    @abstractmethod
    def get_depth_of_cycle(self) -> float:
        """
        Determines the depth of a detected cycle

        Returns
        -------
        float:
            Depth of a detected cycle in p.u.

        """
        pass

    @abstractmethod
    def get_delta_full_equivalent_cycle(self) -> float:
        """
        Determines the delta in full equivalent cycles [0,1]

        Returns
        -------
        float:
            Delta in full equivalent cycles in p.u.

        """
        pass

    @abstractmethod
    def get_crate(self) -> float:
        """
        Determines the mean c-rate of a detected cycle

        Returns
        -------
        float:
            Mean c-rate of a detected cycle in 1/s

        """
        pass

    @abstractmethod
    def get_mean_soc(self) -> float:
        """
        Determines the mean SOC of a detected cycle

        Returns
        -------
        float:
            Mean SOC of a detected cycle in p.u.
        """
        pass

    @abstractmethod
    def get_full_equivalent_cycle(self) -> float:
        """
        Determines the total number of full equivalent cycles

        Returns
        -------
        float:
            Total number of full equivalent cycles

        """
        pass

    @abstractmethod
    def reset(self) -> None:
        """
        resets all values within a degradation model, if battery is replaced

        Returns
        -------

        """
        pass



from configparser import ConfigParser
import pytest
from simses.commons.config.simulation.general import GeneralSimulationConfig
from simses.commons.config.simulation.profile import ProfileConfig
from simses.commons.config.simulation.system import StorageSystemConfig
from simses.system.housing.abstract_housing import Housing
from simses.system.housing.forty_ft_container import FortyFtContainer
from simses.system.housing.twenty_ft_container import TwentyFtContainer
from simses.system.thermal.solar_irradiation.location import \
    LocationSolarIrradiationModel


class TestLocationSolarIrradiationModel:

    none_config = None
    general_config = GeneralSimulationConfig(none_config)
    profile_config = ProfileConfig(none_config)

    temperature: float = 300  # K

    # Ensure if path from simulation.defaults.ini to ambient temperature model works
    solar_irradiation_model_config: ConfigParser = ConfigParser()
    solar_irradiation_model_config.add_section('STORAGE_SYSTEM')
    solar_irradiation_model_config.set('STORAGE_SYSTEM', 'SOLAR_IRRADIATION_MODEL', 'LocationSolarIrradiationModel')
    storage_system_config = StorageSystemConfig(solar_irradiation_model_config)
    solar_irradiation_model = str(storage_system_config.solar_irradiation_model[StorageSystemConfig.SOLAR_IRRADIATION_TYPE])

    # Check compatibility of LocationSolarIrradiationModel with each Housing Model and Ambient Temperature Model
    # ambient_temperature_models = [LocationAmbientTemperature(temperature_config, general_config), ConstantAmbientTemperature()]
    housing_models = list()
    # for model in ambient_temperature_models:
    housing_models.append(TwentyFtContainer(list(), temperature))
    housing_models.append(FortyFtContainer(list(), temperature))

    start_time = 0
    end_time = 100
    sample_time = 1
    time_step_range = range(start_time, end_time, sample_time)

    def create_model(self, housing: Housing):
        if self.solar_irradiation_model == LocationSolarIrradiationModel.__name__:
            return LocationSolarIrradiationModel(self.profile_config, self.general_config, housing)
        else:
            raise Exception('Test failed: Check link between simulation config and factory.')

    @pytest.mark.parametrize("time_step", time_step_range)
    def test_get_heat_load_timestep(self, time_step):
        for housing_type in self.housing_models:
            uut: LocationSolarIrradiationModel = self.create_model(housing_type)
            assert 1400 >= uut.get_global_horizontal_irradiance(time_step) >= 0
            assert (1400 * (2 * (housing_type.outer_layer.surface_area_long_side + housing_type.outer_layer.surface_area_short_side) + housing_type.outer_layer.surface_area_roof)) > uut.get_heat_load(0)




from simses.technology.lithium_ion.cell.format.abstract import CellFormat


class PouchCell(CellFormat):

    def __init__(self, height: float, width: float, length: float):
        super(PouchCell, self).__init__()
        self.__volume = height * width * length  # m3
        self.__surface_area = 2.0 * (length * height + length * width + width * height)  # m2

    def get_volume(self) -> float:
        return self.__volume

    def get_surface_area(self) -> float:
        return self.__surface_area




import numpy as np
import pandas as pd

from simses.commons.config.data.electrolyzer import ElectrolyzerDataConfig
from simses.commons.log import Logger
from simses.commons.state.technology.electrolyzer import ElectrolyzerState
from simses.technology.hydrogen.electrolyzer.stack.alkaline.electrical_model import \
    AlkalineElectricalModel
from simses.technology.hydrogen.electrolyzer.stack.alkaline.fluid_model import \
    AlkalineFluidModel
from simses.technology.hydrogen.electrolyzer.stack.alkaline.parameters import \
    ParametersAlkalineElectrolyzerFit
from simses.technology.hydrogen.electrolyzer.stack.alkaline.pressure_model import \
    AlkalinePressureModel
from simses.technology.hydrogen.electrolyzer.stack.stack_model import ElectrolyzerStackModel


class AlkalineElectrolyzer(ElectrolyzerStackModel):
    """An Alkaline Electrolyzer is a special type of Electrolyzer"""

    # TODO: Move these Parameters to Config, so it is more easily adjustable?
    __NOM_POWER_CELL = 208  # W
    __MAX_POWER_CELL = 208  # W
    __RATIO_NOM_TO_MAX = __NOM_POWER_CELL / __MAX_POWER_CELL

    __NOMINAL_CURRENT_DENSITY = 0.3  # A/cm²
    __MAX_CURRENTDENSITY_CELL = 0.3  # A/cm2
    __NOM_CURRENTDENSITY_CELL = 0.3  # A/cm²
    # __MIN_CURRENTDENSITY_CELL = 0.03  # A/cm2
    __MIN_CURRENTDENSITY_CELL = 0  # A/cm2
    __NOM_VOLTAGE_CELL = 2  # V

    # TODO Find Alkaline Values!
    # __HEAT_CAPACITY = 0.448           # J/K/W
    __HEAT_CAPACITY = 625000 / 26000    # J/K/W (for 26kW Phoebus Electrolyzer)
    __THERMAL_RESISTANCE = 0.167        # K/W (for 26kW Phoebus Electrolyzer)
    __WATER_POWER_DENSITY = 0.125       # kg / kW from: Maximilian Möckl ZAE

    def __init__(self, electrolyzer_maximal_power: float, electrolyzer_data_config: ElectrolyzerDataConfig):
        super().__init__()
        self.__log: Logger = Logger(type(self).__name__)
        self.__number_cells: float = max(round(electrolyzer_maximal_power / self.__MAX_POWER_CELL), 1)
        self.__max_stack_power: float = self.__number_cells * self.__MAX_POWER_CELL  # W
        self.__nominal_stack_power: float = self.__RATIO_NOM_TO_MAX * self.__max_stack_power  # W
        # Scale heat capacity of Phoebus Electrolyser (26kW) to different systems?
        self.__heat_capacity_stack: float = self.__HEAT_CAPACITY * self.__nominal_stack_power  # J/K
        # Scale thermal resistance of Phoebus Electrolyser (26kW) to different systems?
        self.__thermal_resistance_stack: float = self.__THERMAL_RESISTANCE * (26000 / self.__nominal_stack_power)
        self.__water_in_stack: float = self.__WATER_POWER_DENSITY * self.__nominal_stack_power / 1000.0  # 0.125 kg/kW
        self.__log.info('maximal stack power of electrolyzer was adapted to: ' + str(self.__max_stack_power))
        self.__log.info('nominal stack power of electrolyzer is: ' + str(self.__nominal_stack_power))
        self.__log.info('number of serial cells is: ' + str(self.__number_cells))
        # initialize models
        self.__parameters = ParametersAlkalineElectrolyzerFit(electrolyzer_data_config)
        self.__pressure_model: AlkalinePressureModel = AlkalinePressureModel()
        self.__fluid_model: AlkalineFluidModel = AlkalineFluidModel()
        self.__electrical_model: AlkalineElectricalModel = AlkalineElectricalModel(self.__pressure_model,
                                                                                   self.__fluid_model,
                                                                                   electrolyzer_data_config,
                                                                                   self.__parameters)
        # initialize variables
        self.__voltage_stack: float = 0.0
        self.__current_stack: float = 0.0
        self.__current_cell: float = 0.0
        self.__h2o_use_stack: float = 0.0
        self.__o2_generation_stack: float = 0.0
        self.__h2_generation_stack: float = 0.0
        self.__part_pressure_h2: float = 0.0
        self.__part_pressure_o2: float = 0.0
        self.__sat_pressure_h2o: float = 0.0

    def calculate(self, power: float, state: ElectrolyzerState):
        # calculation of current, partial pressures and cell voltage at cell level
        power_cell: float = power / self.__number_cells     # W
        # calculation of current and voltage at stack level
        current_cell = self.__electrical_model.get_current(power_cell, state)  # A
        # TODO is this consideration of operating range valid?
        if (current_cell / (self.__electrical_model.get_geometric_area_cell() * 10000)) < self.__MIN_CURRENTDENSITY_CELL:
            current_cell = 0.0
        voltage_cell = self.__electrical_model.get_cell_voltage(current_cell, state)  # V
        stack_temperature = state.temperature           # °C
        stack_temperature = stack_temperature + 273.15  # K
        self.__current_stack = current_cell  # A  cells are serial connected to a Stack
        self.__current_cell = self.__current_stack
        self.__voltage_stack = voltage_cell * self.__number_cells  # V  cells are serial connected to one stack
        # calculation of pressures TODO more fitting variables for different pressures used in alkaline model
        molarity = self.__fluid_model.get_molarity(stack_temperature)
        self.__part_pressure_h2 = self.__pressure_model.get_partial_pressure_h2(molarity, stack_temperature)
        self.__part_pressure_o2 = self.__pressure_model.get_partial_pressure_o2(molarity, stack_temperature)
        # self.__sat_pressure_h2o = self.__pressure_model.get_sat_pressure_h2o(molarity, stack_temperature)
        # net use of water, production of hydrogen and oxygen at Stack level
        h2_net_cathode = self.__fluid_model.get_h2_generation_cell(self.__current_cell)  # mol/s
        o2_net_anode = self.__fluid_model.get_o2_generation_cell(self.__current_cell)  # mol/s
        h2o_net_use_cell: float = self.__fluid_model.get_h2o_use_cell(self.__current_cell)
        self.__h2o_use_stack = h2o_net_use_cell * self.__number_cells  # mol/s
        self.__o2_generation_stack = o2_net_anode * self.__number_cells  # mol/s
        self.__h2_generation_stack = h2_net_cathode * self.__number_cells  # mol/s

    def get_nominal_current_density(self):
        return self.__NOMINAL_CURRENT_DENSITY

    def get_reference_voltage_eol(self, resistance_increase, exchange_current_decrease) -> float:   # TODO replace placeholder
        return 0

    def get_current(self):
        return self.__current_stack

    def get_current_density(self):
        return self.__current_stack / (self.__electrical_model.get_geometric_area_cell() * 10000)

    def get_hydrogen_production(self):
        return self.__h2_generation_stack

    def get_oxygen_production(self):
        return self.__o2_generation_stack

    def get_voltage(self):
        return self.__voltage_stack

    def get_water_use(self):
        return self.__h2o_use_stack

    def get_number_cells(self):
        return self.__number_cells

    def get_geom_area_stack(self):
        return self.__number_cells * self.__electrical_model.get_geometric_area_cell()

    def get_nominal_stack_power(self):
        return self.__nominal_stack_power

    def get_heat_capacity_stack(self):
        return self.__heat_capacity_stack

    def get_thermal_resistance_stack(self):
        return self.__thermal_resistance_stack

    def get_water_in_stack(self):
        return self.__water_in_stack

    def get_partial_pressure_h2(self):
        return self.__part_pressure_h2

    def get_partial_pressure_o2(self):
        return self.__part_pressure_o2

    def get_sat_pressure_h2o(self):
        return self.__sat_pressure_h2o

    def get_efficiency_curve(self, hydrogen_state: ElectrolyzerState):
        return pd.DataFrame({'current density': pd.Series(np.arange(0, 1, 0.1)), 'cell efficiency': pd.Series(np.arange(0, 1, 0.1)),
                             'voltage efficiency': pd.Series(np.arange(0, 1, 0.1)),
                             'faraday efficiency': pd.Series(np.arange(0, 1, 0.1))})

    def close(self):
        self.__log.close()
